"""RF devices."""

import time as _time
import numpy as _np

from mathphys.functions import get_namedtuple as _get_namedtuple

from .device import DeviceNC as _DeviceNC, Devices as _Devices


class RFGen(_DeviceNC):
    """Wrap the basic features of the RF Generator used in Sirius.

    The Generator is the R&S SMA100A Signal Generator.
    For more informations of the functionalities of this equipment, please
    refer to:
        https://scdn.rohde-schwarz.com/ur/pws/dl_downloads/dl_common_library/dl_manuals/gb_1/s/sma/SMA100A_OperatingManual_en_14.pdf

    This class interacts with an EPICS IOC made for the generator mentioned
    above. More informations on the IOC can be found in:
        https://github.com/lnls-dig/rssmx100a-epics-ioc

    """

    RF_DELTA_MIN = 0.01  # [Hz]
    RF_DELTA_MAX = 15000.0  # [Hz]
    RF_DELTA_RMP = 200  # [Hz]

    PHASE_CONTINUOUS = _get_namedtuple('PhaseContinuous', ('OFF', 'ON'))
    FREQ_OPMODE = _get_namedtuple(
        'FreqOpMode', (
            'CW',  # Continuous wave (regular mode of operation.)
            'SWE',  # Sweep Mode turned on
            'LIST')  # I don't know yet.
        )
    SWEEP_MODE = _get_namedtuple(
        'SweepMode', (
            'AUTO',  # each trigger executes waveform
            'MAN',  # execution is manual
            'STEP')  # each trigger executes one step
        )
    SWEEP_TRIG_SRC = _get_namedtuple(
        'SweepTrigSrc', (
            'AUTO',  # run continuously as soon as mode is configured.
            'SING',  # run single waveform with internal trigger.
            'EXT',   # run single waveform with external trigger.
            'EAUT')  # external trigger tells when to start and stop.
        )
    SWEEP_SPACING = _get_namedtuple(
        'SweepSpacing', (
            'LIN',  # spacing between steps is linear
            'LOG')  # spacing between steps is logarithmic
        )
    SWEEP_SHAPE = _get_namedtuple(
        'SweepShape', (
            'SAWT',  # sawtooth-like waveform
            'TRI')  # triangular waveform
        )
    SWEEP_RETRACE = _get_namedtuple('SweepRetrace', ('OFF', 'ON'))
    SWEEP_RUNNING = _get_namedtuple('SweepRunning', ('OFF', 'ON'))

    class DEVICES:
        """Devices names."""

        AS = 'RF-Gen'
        SPARE = 'RF-SI-Gen'
        ALL = (AS, SPARE)

    _properties = (
        'GeneralFreq-SP', 'GeneralFreq-RB',

        'FreqFExeSweep-Cmd',  # execute single sweep
        'FreqRst-Cmd',  # reset sweeps
        'FreqFRunnMode-Mon',  # is sweep running?

        'GeneralFSweep-Sel', 'GeneralFSweep-Sts',  # frequency opmode
        'FreqFSweepMode-Sel', 'FreqFSweepMode-Sts',  # sweep mode
        'TrigFSweepSrc-Sel', 'TrigFSweepSrc-Sts',  # trigger source
        'FreqFSpacMode-Sel', 'FreqFSpacMode-Sts',  # spacing mode
        'FreqFreqRetr-Sel', 'FreqFreqRetr-Sts',  # retrace
        'FreqFreqShp-Sel', 'FreqFreqShp-Sts',  # shape
        'FreqPhsCont-Sts',  # 'FreqPhsCont-Sel', (setpoint is not needed here)

        'FreqFreqSpan-RB', 'FreqFreqSpan-SP',
        'FreqCenterFreq-RB', 'FreqCenterFreq-SP',
        'FreqStartFreq-RB', 'FreqStartFreq-SP',
        'FreqStopFreq-RB', 'FreqStopFreq-SP',
        'FreqFStepLin-RB', 'FreqFStepLin-SP',
        'FreqFStepLog-RB', 'FreqFStepLog-SP',
        'FreqFDwellTime-RB', 'FreqFDwellTime-SP',
        )

    def __init__(self, devname=None):
        """."""
        if devname is None:
            devname = RFGen.DEVICES.AS

        # check if device exists
        if devname not in RFGen.DEVICES.ALL:
            raise NotImplementedError(devname)

        # call base class constructor
        super().__init__(devname, properties=RFGen._properties)

    @property
    def is_phase_continuous(self):
        """."""
        return self['FreqPhsCont-Sts'] == self.PHASE_CONTINUOUS.ON

    @property
    def frequency(self):
        """."""
        return self['GeneralFreq-RB']

    @frequency.setter
    def frequency(self, value):
        delta_max = RFGen.RF_DELTA_RMP  # [Hz]
        freq0 = self.frequency
        if freq0 is None or value is None:
            return
        delta = abs(value-freq0)
        if delta < RFGen.RF_DELTA_MIN or delta > RFGen.RF_DELTA_MAX:
            return
        npoints = int(delta/delta_max) + 2
        freq_span = _np.linspace(freq0, value, npoints)[1:]
        self._pvs['GeneralFreq-SP'].put(freq_span[0], wait=False)
        for freq in freq_span[1:]:
            _time.sleep(1.0)
            self._pvs['GeneralFreq-SP'].put(freq, wait=False)
        self['GeneralFreq-SP'] = value

    def set_frequency(self, value, tol=1, timeout=10):
        """Set RF phase and wait until it gets there."""
        self.frequency = value
        return self._wait_float(
            'GeneralFreq-RB', value, abs_tol=tol, timeout=timeout)

    @property
    def freq_opmode(self):
        """Operation mode of the frequency generator.

        Takes the following values:
            0 - 'CW':  Continuous wave (regular mode of operation.)
            1 - 'SWE': Seep Mode turned on.
            2 - 'LIST': I don't know yet.

        """
        return self['GeneralFSweep-Sts']

    @freq_opmode.setter
    def freq_opmode(self, value):
        self._enum_setter('GeneralFSweep-Sel', value, self.FREQ_OPMODE)

    @property
    def freq_opmode_str(self):
        """Operation mode of the frequency generator.

        Takes the following values:
            0 - 'CW':  Continuous wave (regular mode of operation.)
            1 - 'SWE': Seep Mode turned on.
            2 - 'LIST': I don't know yet.

        """
        return self.FREQ_OPMODE._fields[self['GeneralFSweep-Sts']]

    def cmd_freq_opmode_to_continuous_wave(self):
        """."""
        self.freq_opmode = self.FREQ_OPMODE.CW
        return self._wait('GeneralFSweep-Sts', self.FREQ_OPMODE.CW)

    def cmd_freq_opmode_to_sweep(self):
        """."""
        self.freq_opmode = self.FREQ_OPMODE.SWE
        return self._wait('GeneralFSweep-Sts', self.FREQ_OPMODE.SWE)

    @property
    def freq_sweep_is_running(self):
        """Check whether any frequency sweep is in progress."""
        return self['FreqFRunnMode-Mon'] == self.SWEEP_RUNNING.ON

    def wait_freq_sweep_to_finish(self, timeout=10):
        """Wait until the current frequency sweep is finished."""
        return self._wait(
            'FreqFRunnMode-Mon', self.SWEEP_RUNNING.OFF, timeout=timeout)

    def cmd_freq_sweep_start(self):
        """Send an internal trigger to start frequency sweep.

        Only effective when `freq_sweep_trig_src` is in "SING".

        """
        self['FreqFExeSweep-Cmd'] = 1
        return True

    def cmd_freq_sweep_reset(self):
        """Reset current frequency sweep and wait to start new one."""
        self['FreqRst-Cmd'] = 1
        return True

    @property
    def freq_sweep_mode(self):
        """Mode of the frequency sweep.

        Takes the following values:
            0 - 'AUTO': each trigger executes the waveform;
            1 - 'MAN': execution is manual;
            2 - 'STEP': each trigger executes one step of the waveform.

        """
        return self['FreqFSweepMode-Sts']

    @freq_sweep_mode.setter
    def freq_sweep_mode(self, value):
        self._enum_setter('FreqFSweepMode-Sel', value, self.SWEEP_MODE)

    @property
    def freq_sweep_mode_str(self):
        """Mode of the frequency sweep.

        Takes the following values:
            0 - 'AUTO': each trigger executes the waveform;
            1 - 'MAN': execution is manual;
            2 - 'STEP': each trigger executes one step of the waveform.

        """
        return self.SWEEP_MODE._fields[self['FreqFSweepMode-Sts']]

    def cmd_freq_sweep_mode_to_automatic(self):
        """."""
        self.freq_sweep_mode = self.SWEEP_MODE.AUTO
        return self._wait('FreqFSweepMode-Sts', self.SWEEP_MODE.AUTO)

    def cmd_freq_sweep_mode_to_step(self):
        """."""
        self.freq_sweep_mode = self.SWEEP_MODE.STEP
        return self._wait('FreqFSweepMode-Sts', self.SWEEP_MODE.STEP)

    @property
    def freq_sweep_spacing_mode(self):
        """Define the spacing scale between steps of the sweep.

        Takes the following values:
            0 - 'LIN': Constant spaces between steps;
            1 - 'LOG': Logarithmic spaces between steps;

        """
        return self['FreqFSpacMode-Sts']

    @freq_sweep_spacing_mode.setter
    def freq_sweep_spacing_mode(self, value):
        self._enum_setter('FreqFSpacMode-Sel', value, self.SWEEP_SPACING)

    @property
    def freq_sweep_spacing_mode_str(self):
        """Define the spacing scale between steps of the sweep.

        Takes the following values:
            0 - 'LIN': Constant spaces between steps;
            1 - 'LOG': Logarithmic spaces between steps;

        """
        return self.SWEEP_SPACING._fields[self['FreqFSpacMode-Sts']]

    def cmd_freq_sweep_spacing_mode_to_linear(self):
        """."""
        self.freq_sweep_spacing_mode = self.SWEEP_SPACING.LIN
        return self._wait('FreqFSpacMode-Sts', self.SWEEP_SPACING.LIN)

    def cmd_freq_sweep_spacing_mode_to_log(self):
        """."""
        self.freq_sweep_spacing_mode = self.SWEEP_SPACING.LOG
        return self._wait('FreqFSpacMode-Sts', self.SWEEP_SPACING.LOG)

    @property
    def freq_sweep_trig_src(self):
        """Define the trigger source for the sweep.

        Takes the following values:
            0 - 'AUTO': run continuously as soon as `freq_sweep_mode` is
                    changed to `SWE`.
            1 - 'SING': run single waveform with internal trigger.
            2 - 'EXT': run single waveform with external trigger.
            3 - 'EAUT': external trigger tells when to start and stop.

        """
        return self['TrigFSweepSrc-Sts']

    @freq_sweep_trig_src.setter
    def freq_sweep_trig_src(self, value):
        self._enum_setter('TrigFSweepSrc-Sel', value, self.SWEEP_TRIG_SRC)

    @property
    def freq_sweep_trig_src_str(self):
        """Define the trigger source for the sweep.

        Takes the following values:
            0 - 'AUTO': run continuously as soon as `freq_sweep_mode` is
                    changed to `SWE`.
            1 - 'SING': run single waveform with internal trigger.
            2 - 'EXT': run single waveform with external trigger.
            3 - 'EAUT': external trigger tells when to start and stop.

        """
        return self.SWEEP_TRIG_SRC._fields[self['TrigFSweepSrc-Sts']]

    def cmd_freq_sweep_trig_src_to_external(self):
        """."""
        self.freq_sweep_trig_src = self.SWEEP_TRIG_SRC.EXT
        return self._wait('TrigFSweepSrc-Sts', self.SWEEP_TRIG_SRC.EXT)

    def cmd_freq_sweep_trig_src_to_single(self):
        """."""
        self.freq_sweep_trig_src = self.SWEEP_TRIG_SRC.SING
        return self._wait('TrigFSweepSrc-Sts', self.SWEEP_TRIG_SRC.SING)

    def cmd_freq_sweep_trig_src_to_auto(self):
        """."""
        self.freq_sweep_trig_src = self.SWEEP_TRIG_SRC.AUTO
        return self._wait('TrigFSweepSrc-Sts', self.SWEEP_TRIG_SRC.AUTO)

    @property
    def freq_sweep_shape(self):
        """Define the shape of the sweep.

        Takes the following values:
            0 - 'SAWT': sawtooth-like waveform:
                                _           _           _
                              _| |        _| |        _| |
                            _|   |      _|   |      _|   |
                -----|    _|     |    _|     |    _|     |-----
                     |  _|       |  _|       |  _|
                     |_|         |_|         |_|
            1 - 'TRI': triangular waveform:
                                _                   _
                              _| |_               _| |_
                            _|     |_           _|     |_
                -----|    _|         |_       _|         |_    |-----
                     |  _|             |_   _|             |_  |
                     |_|                 |_|                 |_|

        """
        return self['FreqFreqShp-Sts']

    @freq_sweep_shape.setter
    def freq_sweep_shape(self, value):
        self._enum_setter('FreqFreqShp-Sel', value, self.SWEEP_SHAPE)

    @property
    def freq_sweep_shape_str(self):
        """Define the shape of the sweep.

        Takes the following values:
            0 - 'SAWT': sawtooth-like waveform:
                                _           _           _
                              _| |        _| |        _| |
                            _|   |      _|   |      _|   |
                -----|    _|     |    _|     |    _|     |-----
                     |  _|       |  _|       |  _|
                     |_|         |_|         |_|
            1 - 'TRI': triangular waveform:
                                _                   _
                              _| |_               _| |_
                            _|     |_           _|     |_
                -----|    _|         |_       _|         |_    |-----
                     |  _|             |_   _|             |_  |
                     |_|                 |_|                 |_|

        """
        return self.SWEEP_SHAPE._fields[self['FreqFreqShp-Sts']]

    def cmd_freq_sweep_shape_to_sawtooth(self):
        """."""
        self.freq_sweep_shape = self.SWEEP_SHAPE.SAWT
        return self._wait('FreqFreqShp-Sts', self.SWEEP_SHAPE.SAWT)

    def cmd_freq_sweep_shape_to_triangular(self):
        """."""
        self.freq_sweep_shape = self.SWEEP_SHAPE.TRI
        return self._wait('FreqFreqShp-Sts', self.SWEEP_SHAPE.TRI)

    @property
    def freq_sweep_retrace(self):
        """Define whether to return to start freq. after sweep finish.

        Takes the following values:
            0 - 'OFF': return to starting frequency.
            1 - 'ON': stay at last frequency of the sweep.

        """
        return self['FreqFreqRetr-Sts']

    @freq_sweep_retrace.setter
    def freq_sweep_retrace(self, value):
        self._enum_setter('FreqFreqRetr-Sel', value, self.SWEEP_RETRACE)

    @property
    def freq_sweep_retrace_str(self):
        """Define whether to return to start freq. after sweep finish.

        Takes the following values:
            0 - 'OFF': return to starting frequency.
            1 - 'ON': stay at last frequency of the sweep.

        """
        return self.SWEEP_RETRACE._fields[self['FreqFreqRetr-Sts']]

    def cmd_freq_sweep_retrace_turn_off(self):
        """."""
        self.freq_sweep_retrace = self.SWEEP_RETRACE.OFF
        return self._wait('FreqFreqRetr-Sts', self.SWEEP_RETRACE.OFF)

    def cmd_freq_sweep_retrace_turn_on(self):
        """."""
        self.freq_sweep_retrace = self.SWEEP_RETRACE.ON
        return self._wait('FreqFreqRetr-Sts', self.SWEEP_RETRACE.ON)

    @property
    def freq_sweep_span(self):
        """Total span of the frequency sweep in [Hz]."""
        return self['FreqFreqSpan-RB']

    @freq_sweep_span.setter
    def freq_sweep_span(self, value):
        self['FreqFreqSpan-SP'] = float(value)

    @property
    def freq_sweep_center_freq(self):
        """Center frequency of the frequency sweep in [Hz]."""
        return self['FreqCenterFreq-RB']

    @freq_sweep_center_freq.setter
    def freq_sweep_center_freq(self, value):
        self['FreqCenterFreq-SP'] = float(value)

    @property
    def freq_sweep_start_freq(self):
        """Start point of the frequency sweep in [Hz]."""
        return self['FreqStartFreq-RB']

    @freq_sweep_start_freq.setter
    def freq_sweep_start_freq(self, value):
        self['FreqStartFreq-SP'] = float(value)

    @property
    def freq_sweep_stop_freq(self):
        """Other extreme of the frequency sweep in [Hz]."""
        return self['FreqStopFreq-RB']

    @freq_sweep_stop_freq.setter
    def freq_sweep_stop_freq(self, value):
        self['FreqStopFreq-SP'] = float(value)

    @property
    def freq_sweep_step_freq_linear(self):
        """Step used if `freq_sweep_spacing_mode` is linear, in [Hz]."""
        return self['FreqFStepLin-RB']

    @freq_sweep_step_freq_linear.setter
    def freq_sweep_step_freq_linear(self, value):
        self['FreqFStepLin-SP'] = float(value)

    @property
    def freq_sweep_step_freq_log(self):
        """Step used if `freq_sweep_spacing_mode` is linear, in [%]."""
        return self['FreqFStepLog-RB']

    @freq_sweep_step_freq_log.setter
    def freq_sweep_step_freq_log(self, value):
        self['FreqFStepLog-SP'] = float(value)

    @property
    def freq_sweep_step_time(self):
        """Time to stay in each step of the sweep in [s]."""
        return self['FreqFDwellTime-RB']

    @freq_sweep_step_time.setter
    def freq_sweep_step_time(self, value):
        self['FreqFDwellTime-SP'] = float(value)

    def configure_freq_sweep(
            self, ampl, nr_steps, duration, sawtooth=True, centered=True,
            increasing=True, retrace=False):
        """Configure generator to create a single frequency sweep.

        Currently the sweep will be ready to be triggered by an internal
        trigger (see method `cmd_freq_sweep_start`).

        Args:
            ampl (float): Amplitude of the sweep in Hz. NOTE: Be careful not
                to create discontinuities larger than 200 Hz in the frequency,
                which could make the LLRF lose track of the reference
                frequency.
            nr_steps (int): number of steps to divide the ramp. Notice that if
                the triangular waveform is chosen, the actual number of steps
                will be 2 times minus 1 the value provided here. Must be
                larger than 1, so stepsize is at least equal to span.
            duration (float): total desired duration in seconds.
            sawtooth (bool, optional): If True, then the waveform will be se to
                sawtooth. If False, the triangular waveform will be chosen.
                Defaults to True.
            centered (bool, optional): If True the current frequency will be
                centered in the ramp. If False, it will be equal to the
                starting frequency. Defaults to True.
            increasing (bool, optional): Whether the ramp will start at the
                smaller frequency and go to the higher one (True) or
                vice-versa (False). Defaults to True.
            retrace (bool, optional): Whether (True) or not (False) to go back
                to the current frequency after finish the sweep.
                Detaults to False.

        Returns:
            ampl (float): real amplitude that will be used in sweep.

        Raises:
            ValueError: raised if nr_steps is <= 1.

        """
        ampl = abs(ampl)
        nr_steps = int(nr_steps)
        if nr_steps <= 1:
            raise ValueError('Input nr_steps must be larger than 1.')
        nr_interv = nr_steps-1

        span = 2*ampl
        stepsize = round(span / nr_interv, 2)
        span = nr_interv * stepsize
        span *= 1 if increasing else -1
        stepsize *= 1 if increasing else -1
        curr_freq = self.frequency
        ampl = span/2

        if sawtooth:
            self.cmd_freq_sweep_shape_to_sawtooth()
            dwelltime = duration / nr_steps
        else:
            self.cmd_freq_sweep_shape_to_triangular()
            dwelltime = duration / (2*nr_steps - 1)
        self.freq_sweep_step_time = dwelltime

        # NOTE: Depending whether the sweep is centered or not it is easier to
        # set the center_freq and span or start and stop
        if centered:
            self.freq_sweep_center_freq = curr_freq
            self.freq_sweep_span = span
        else:
            self.freq_sweep_start_freq = curr_freq
            self.freq_sweep_stop_freq = curr_freq + span

        if retrace:
            self.cmd_freq_sweep_retrace_turn_on()
        else:
            self.cmd_freq_sweep_retrace_turn_off()

        self.cmd_freq_sweep_spacing_mode_to_linear()
        self.freq_sweep_step_freq_linear = stepsize

        self.cmd_freq_sweep_mode_to_automatic()
        self.cmd_freq_sweep_trig_src_to_single()

        # NOTE: do not change freq_opmode here, because this will make the
        #  generator automatically go to the starting frequency, regardless of
        # the trigger type:
        # self.cmd_freq_opmode_to_sweep()
        return ampl


class ASLLRF(_DeviceNC):
    """AS LLRF."""
    VoltIncRates = _get_namedtuple('VoltIncRates', (
        'vel_0p01', 'vel_0p03', 'vel_0p1', 'vel_0p25', 'vel_0p5', 'vel_1p0',
        'vel_2p0', 'vel_4p0', 'vel_6p0', 'vel_8p0', 'vel_10p0', 'vel_15p0',
        'vel_20p0', 'vel_30p0', 'vel_50p0', 'Immediately'))
    PhsIncRates = _get_namedtuple('PhsIncRates', (
        'vel_0p1', 'vel_0p2', 'vel_0p5', 'vel_1p0', 'vel_2p0', 'vel_5p0',
        'vel_10p0', 'Immediately'))

    class DEVICES:
        """Devices names."""

        BO = 'BR-RF-DLLRF-01'
        SI = 'SR-RF-DLLRF-01'
        ALL = (BO, SI)

    _properties = (
        'PL:REF', 'PL:REF:S', 'SL:REF:PHS', 'SL:INP:PHS',
        'mV:AL:REF-SP', 'mV:AL:REF-RB', 'SL:REF:AMP', 'SL:INP:AMP',
        'DTune-SP', 'DTune-RB', 'TUNE:DEPHS',
        'RmpPhsBot-SP', 'RmpPhsBot-RB', 'RmpPhsTop-SP', 'RmpPhsTop-RB',
        'RmpEnbl-Sel', 'RmpEnbl-Sts', 'RmpReady-Mon',
        'FF:ON', 'FF', 'FF:S', 'FF:POS', 'FF:POS:S',
        'FF:GAIN:CELL2', 'FF:GAIN:CELL2:S', 'FF:GAIN:CELL4', 'FF:GAIN:CELL4:S',
        'FF:DEADBAND', 'FF:DEADBAND:S', 'FF:CELL2', 'FF:CELL4', 'FF:ERR',
        'AMPREF:INCRATE', 'AMPREF:INCRATE:S',
        'PHSREF:INCRATE', 'PHSREF:INCRATE:S',
        )

    def __init__(self, devname):
        """."""
        # check if device exists
        if devname not in ASLLRF.DEVICES.ALL:
            raise NotImplementedError(devname)

        # call base class constructor
        super().__init__(devname, properties=ASLLRF._properties)

    @property
    def is_cw(self):
        """Is CW."""
        return not self.rmp_enable

    @property
    def rmp_enable(self):
        """Ramp enable."""
        return self['RmpEnbl-Sts']

    @rmp_enable.setter
    def rmp_enable(self, value):
        self['RmpEnbl-Sel'] = value

    @property
    def rmp_ready(self):
        """Ramp ready."""
        return self['RmpReady-Mon']

    @property
    def phase_top(self):
        """."""
        return self['RmpPhsTop-RB']

    @phase_top.setter
    def phase_top(self, value):
        self['RmpPhsTop-SP'] = self._wrap_phase(value)

    @property
    def phase_bottom(self):
        """."""
        return self['RmpPhsBot-RB']

    @phase_bottom.setter
    def phase_bottom(self, value):
        self['RmpPhsBot-SP'] = self._wrap_phase(value)

    @property
    def phase_mon(self):
        """."""
        return self['SL:INP:PHS']

    @property
    def phase_ref(self):
        """."""
        return self['SL:REF:PHS']

    @property
    def phase(self):
        """."""
        return self['PL:REF']

    @phase.setter
    def phase(self, value):
        self['PL:REF:S'] = self._wrap_phase(value)

    @property
    def phase_incrate_str(self):
        """."""
        return self.PhsIncRates._fields[self['PHSREF:INCRATE']]

    @property
    def phase_incrate(self):
        """."""
        return self['PHSREF:INCRATE']

    @phase_incrate.setter
    def phase_incrate(self, value):
        if int(value) in self.PhsIncRates:
            self['PHSREF:INCRATE:S'] = int(value)

    def set_phase(self, value, tol=0.2, timeout=10, wait_mon=False):
        """Set RF phase and wait until it gets there."""
        self.phase = value
        pv2wait = 'SL:INP:PHS' if wait_mon else 'SL:REF:PHS'
        return self._wait_float(pv2wait, value, abs_tol=tol, timeout=timeout)

    @property
    def voltage_mon(self):
        """."""
        return self['SL:INP:AMP']

    @property
    def voltage_ref(self):
        """."""
        return self['SL:REF:AMP']

    @property
    def voltage(self):
        """."""
        return self['mV:AL:REF-RB']

    @voltage.setter
    def voltage(self, value):
        self['mV:AL:REF-SP'] = value

    @property
    def voltage_incrate_str(self):
        """."""
        return self.VoltIncRates._fields[self['AMPREF:INCRATE']]

    @property
    def voltage_incrate(self):
        """."""
        return self['AMPREF:INCRATE']

    @voltage_incrate.setter
    def voltage_incrate(self, value):
        if int(value) in self.VoltIncRates:
            self['AMPREF:INCRATE:S'] = int(value)

    def set_voltage(self, value, tol=1, timeout=10, wait_mon=False):
        """Set RF phase and wait until it gets there."""
        self.voltage = value
        pv2wait = 'SL:INP:AMP' if wait_mon else 'SL:REF:AMP'
        return self._wait_float(pv2wait, value, abs_tol=tol, timeout=timeout)

    @property
    def detune(self):
        """."""
        return self['DTune-RB']

    @detune.setter
    def detune(self, value):
        self['DTune-SP'] = value

    @property
    def detune_error(self):
        """."""
        return self['TUNE:DEPHS']

    @property
    def field_flatness_acting(self):
        """Return whether the field flatness loop is acting."""
        return self['FF:ON']

    @property
    def field_flatness_enable(self):
        """Return whether the field flatness loop is enabled."""
        return self['FF']

    @field_flatness_enable.setter
    def field_flatness_enable(self, value):
        """Control state of field flatness loop."""
        self['FF:S'] = bool(value)

    @property
    def field_flatness_position(self):
        """Return the field flatness position."""
        return self['FF:POS']

    @field_flatness_position.setter
    def field_flatness_position(self, value):
        """Control the field flatness position."""
        self['FF:POS:S'] = bool(value)

    @property
    def field_flatness_gain1(self):
        """Return the gain of the first cell controlled."""
        return self['FF:GAIN:CELL2']

    @field_flatness_gain1.setter
    def field_flatness_gain1(self, value):
        """Control the gain of the first cell controlled."""
        self['FF:GAIN:CELL2:S'] = value

    @property
    def field_flatness_gain2(self):
        """Return the gain of the second cell controle."""
        return self['FF:GAIN:CELL4']

    @field_flatness_gain2.setter
    def field_flatness_gain2(self, value):
        """Control the gain of the second cell controlled."""
        self['FF:GAIN:CELL4:S'] = value

    @property
    def field_flatness_deadband(self):
        """Return the loop action deadband in [%]."""
        return self['FF:DEADBAND']

    @field_flatness_deadband.setter
    def field_flatness_deadband(self, value):
        """Control the loop action deadband in [%]."""
        self['FF:DEADBAND:S'] = value

    @property
    def field_flatness_amp1(self):
        """Amplitude of the first cell controled in [mV]."""
        return self['FF:CELL2']

    @property
    def field_flatness_amp2(self):
        """Amplitude of the second cell controled in [mV]."""
        return self['FF:CELL4']

    @property
    def field_flatness_error(self):
        """Return the amplitude error in [mV]."""
        return self['FF:ERR']

    @staticmethod
    def _wrap_phase(phase):
        """Phase must be in [-180, +180] interval."""
        return (phase + 180) % 360 - 180


class BORFCavMonitor(_DeviceNC):
    """."""

    class DEVICES:
        """Devices names."""

        BO = 'BO-05D:RF-P5Cav'

    _properties = (
        'Cell3PwrTop-Mon', 'Cell3PwrBot-Mon', 'PwrRFIntlk-Mon', 'Sts-Mon',
        'Cell1Pwr-Mon', 'Cell2Pwr-Mon', 'Cell3Pwr-Mon', 'Cell4Pwr-Mon',
        'Cell5Pwr-Mon', 'Cylin1T-Mon', 'Cylin2T-Mon', 'Cylin3T-Mon',
        'Cylin4T-Mon', 'Cylin5T-Mon', 'CoupT-Mon',
        'RmpAmpVCavBot-Mon', 'RmpAmpVCavTop-Mon',
        )

    def __init__(self):
        """."""
        # call base class constructor
        super().__init__(
            BORFCavMonitor.DEVICES.BO, properties=BORFCavMonitor._properties)

    @property
    def status(self):
        """."""
        return self['Sts-Mon']

    @property
    def power_interlock(self):
        """."""
        return self['PwrRFIntlk-Mon']

    @property
    def power_top(self):
        """."""
        return self['Cell3PwrTop-Mon']

    @property
    def power_bottom(self):
        """."""
        return self['Cell3PwrBot-Mon']

    @property
    def power_reverse(self):
        """."""
        return self['PwrRev-Mon']

    @property
    def power_forward(self):
        """."""
        return self['PwrFwd-Mon']

    @property
    def power_cell1(self):
        """."""
        return self['Cell1Pwr-Mon']

    @property
    def power_cell2(self):
        """."""
        return self['Cell2Pwr-Mon']

    @property
    def power_cell3(self):
        """."""
        return self['Cell3Pwr-Mon']

    @property
    def power_cell4(self):
        """."""
        return self['Cell4Pwr-Mon']

    @property
    def power_cell5(self):
        """."""
        return self['Cell5Pwr-Mon']

    @property
    def temp_coupler(self):
        """."""
        return self['CoupT-Mon']

    @property
    def temp_cell1(self):
        """."""
        return self['Cylin1T-Mon']

    @property
    def temp_cell2(self):
        """."""
        return self['Cylin2T-Mon']

    @property
    def temp_cell3(self):
        """."""
        return self['Cylin3T-Mon']

    @property
    def temp_cell4(self):
        """."""
        return self['Cylin4T-Mon']

    @property
    def temp_cell5(self):
        """."""
        return self['Cylin5T-Mon']

    @property
    def gap_voltage_bottom(self):
        """Gap Voltage in [V]."""
        return self['RmpAmpVCavBot-Mon']

    @property
    def gap_voltage_top(self):
        """Gap Voltage in [V]."""
        return self['RmpAmpVCavTop-Mon']


class SIRFCavMonitor(_DeviceNC):
    """."""

    class DEVICES:
        """Devices names."""

        SI = 'SI-02SB:RF-P7Cav'

    _properties = (
        'PwrRev-Mon', 'PwrFwd-Mon',
        'PwrCell4Top-Mon', 'PwrCell4Bot-Mon', 'PwrRFIntlk-Mon', 'Sts-Mon',
        'PwrCell2-Mon', 'PwrCell4-Mon', 'PwrCell6-Mon', 'Cylin1T-Mon',
        'Cylin2T-Mon', 'Cylin3T-Mon', 'Cylin4T-Mon', 'Cylin5T-Mon',
        'Cylin6T-Mon', 'Cylin7T-Mon', 'CoupT-Mon', 'AmpVCav-Mon',
        )

    def __init__(self):
        """."""
        # call base class constructor
        super().__init__(
            SIRFCavMonitor.DEVICES.SI, properties=SIRFCavMonitor._properties)

    @property
    def status(self):
        """."""
        return self['Sts-Mon']

    @property
    def power_interlock(self):
        """."""
        return self['PwrRFIntlk-Mon']

    @property
    def power_top(self):
        """."""
        return self['PwrCell4Top-Mon']

    @property
    def power_bottom(self):
        """."""
        return self['PwrCell4Bot-Mon']

    @property
    def power_reverse(self):
        """."""
        return self['PwrRev-Mon']

    @property
    def power_forward(self):
        """."""
        return self['PwrFwd-Mon']

    @property
    def power_cell2(self):
        """."""
        return self['PwrCell2-Mon']

    @property
    def power_cell4(self):
        """."""
        return self['PwrCell4-Mon']

    @property
    def power_cell6(self):
        """."""
        return self['PwrCell6-Mon']

    @property
    def temp_coupler(self):
        """."""
        return self['CoupT-Mon']

    @property
    def temp_cell1(self):
        """."""
        return self['Cylin1T-Mon']

    @property
    def temp_cell2(self):
        """."""
        return self['Cylin2T-Mon']

    @property
    def temp_cell3(self):
        """."""
        return self['Cylin3T-Mon']

    @property
    def temp_cell4(self):
        """."""
        return self['Cylin4T-Mon']

    @property
    def temp_cell5(self):
        """."""
        return self['Cylin5T-Mon']

    @property
    def temp_cell6(self):
        """."""
        return self['Cylin6T-Mon']

    @property
    def temp_cell7(self):
        """."""
        return self['Cylin7T-Mon']

    @property
    def gap_voltage(self):
        """."""
        return self['AmpVCav-Mon']


class RFCav(_Devices):
    """."""

    class DEVICES:
        """Devices names."""

        BO = 'BO-05D:RF-P5Cav'
        SI = 'SI-02SB:RF-P7Cav'
        ALL = (BO, SI)

    def __init__(self, devname):
        """."""
        # check if device exists
        if devname not in RFCav.DEVICES.ALL:
            raise NotImplementedError(devname)

        self.dev_rfgen = RFGen()
        if devname == RFCav.DEVICES.SI:
            self.dev_llrf = ASLLRF(ASLLRF.DEVICES.SI)
            self.dev_cavmon = SIRFCavMonitor()
        elif devname == RFCav.DEVICES.BO:
            self.dev_llrf = ASLLRF(ASLLRF.DEVICES.BO)
            self.dev_cavmon = BORFCavMonitor()
        devices = (self.dev_rfgen, self.dev_llrf, self.dev_cavmon)

        # call base class constructor
        super().__init__(devname, devices)

    @property
    def is_cw(self):
        """."""
        return self.dev_llrf.is_cw

    @property
    def power(self):
        """."""
        return self.dev_cavmon.get_power(self.is_cw)

    def set_voltage(self, value, timeout=10, tol=1, wait_mon=False):
        """."""
        return self.dev_llrf.set_voltage(
            value, tol=tol, timeout=timeout, wait_mon=wait_mon)

    def set_phase(self, value, timeout=10, tol=0.2, wait_mon=False):
        """."""
        return self.dev_llrf.set_phase(
            value, tol=tol, timeout=timeout, wait_mon=wait_mon)

    def set_frequency(self, value, timeout=10, tol=0.05):
        """."""
        return self.dev_rfgen.set_frequency(value, tol=tol, timeout=timeout)


class RFKillBeam(ASLLRF):
    """RF Kill Beam Button."""
    TIMEOUT_WAIT = 20.0  # [s]
    INCRATE_VALUE = ASLLRF.VoltIncRates.vel_50p0  # [mV/s]
    REFMIN_VALUE = 60  # Minimum Amplitude Reference [mV]

    def __init__(self):
        """Init."""
        super().__init__(ASLLRF.DEVICES.SI)

    def cmd_kill_beam(self):
        """Kill beam."""
        if not self.wait_for_connection(self.TIMEOUT_WAIT):
            return [False, 'Could not read RF PVs.']

        # get initial values
        amp_incrate_init = self.voltage_incrate
        amp_init = self.voltage

        # set Amplitude Increase Rate to 50 mV/s and wait
        self.voltage_incrate = self.INCRATE_VALUE
        if not self._wait(
                'AMPREF:INCRATE', self.INCRATE_VALUE,
                timeout=self.TIMEOUT_WAIT):
            return [False, 'Could not set RF Amplitude Increase Rate.']

        # set Amplitude Reference to 60mV and wait
        self.voltage = self.REFMIN_VALUE
        if not self._wait_float(
                'SL:INP:AMP', self.REFMIN_VALUE, abs_tol=1,
                timeout=self.TIMEOUT_WAIT):
            return [False, 'Could not set RF Voltage to low value.']

        # set Amplitude Reference to initial value
        self.voltage = amp_init
        if not self._wait_float(
                'SL:INP:AMP', amp_init, abs_tol=1, timeout=self.TIMEOUT_WAIT):
            return [False, 'Could not set RF Voltage back to original value.']

        # set Amplitude Increase Rate to initial value
        self.voltage_incrate = amp_incrate_init
        if not self._wait(
                'AMPREF:INCRATE', self.INCRATE_VALUE,
                timeout=self.TIMEOUT_WAIT):
            return [False, 'Could not set RF Amplitude Increase Rate back.']
        return [True, '']



"""PRUController.

This module implements classes that are used to do low level BeagleBone
communications, be it with PRU or BSMP requests to power supply controllers
at the other end of the serial line.
"""

from time import (time as _time, sleep as _sleep)
from socket import timeout as _socket_timeout
from copy import deepcopy as _dcopy
from threading import Thread as _Thread
from threading import Lock as _Lock

from ...util import get_timestamp as _get_timestamp
from ...bsmp import SerialError as _SerialError

from ..bsmp.constants import _const_bsmp
from ..bsmp.constants import __version__ as _firmware_version_siriuspy

from .udc import UDC as _UDC
from .psdevstate import PSDevState as _PSDevState


class PRUController:
    """Beaglebone controller.

    This class implements all basic PRU configuration and BSMP communications
    of the Beaglebone computer connected through a serial line to power supply
    controllers.
    """

    # NOTE: All private methods starting with '_bsmp' string invoke serial
    #       bsmp communications.

    # --- public interface ---

    def __init__(self,
                 pru,
                 prucqueue,
                 psmodel,
                 devices,
                 processing=False,
                 scanning=False,
                 freq=None,
                 init=True):
        """Init."""
        # --- Init structures ---

        print()
        print('PRUController: struct initialization')
        print('devices: {}'.format(devices))

        # init timetsamp
        self._timestamp_update = _time()

        # init time interval
        t0_ = _time()

        # init sofb mode to false
        self._sofb_mode = False

        # index of device in self._device_ids for next update in SOFB mode
        self._sofb_update_dev_idx = 0  # cyclical updates!

        # create lock
        self._lock = _Lock()

        # PRU communication object
        self._pru = pru

        # store power supply model
        self._psmodel = psmodel

        # devices
        self._devices = _dcopy(devices)

        # sorted list of device ids
        self._device_ids = sorted([dev[1] for dev in devices])

        # initialize UDC
        self._udc, self._parms, self._psupplies = PRUController._init_udc(
            pru, self._psmodel.name, self._device_ids, freq)

        # index of device in self._device_ids for wfmref/scope update
        self._scope_update = False
        self._scope_update_dev_idx = 0  # cyclical updates!

        # update time interval attribute
        self._scan_interval = self._get_scan_interval()

        # time interval
        t1_ = _time()
        print('TIMING struct init [{:.3f} ms]'.format(1000*(t1_ - t0_)))

        # attributes that control processing flow
        self._queue = prucqueue
        self._processing = processing
        self._scanning = scanning

        # starts communications
        self._dev_idx_last_scanned = None
        self._thread_scan = None
        if init:
            self.bsmp_init_communication()

    # --- properties to read and set controller state and access functions ---

    @property
    def scan_interval(self):
        """Scan interval."""
        return self._scan_interval

    @property
    def scanning(self):
        """Return scanning state."""
        return self._scanning

    @scanning.setter
    def scanning(self, value):
        """Set scanning state."""
        self._scanning = value

    @property
    def processing(self):
        """Return processing state."""
        return self._queue.is_running

    @processing.setter
    def processing(self, value):
        """Set processing state."""
        self._processing = value
        if self._processing:
            self._queue.start()
        else:
            self._queue.stop()

    @property
    def queue_length(self):
        """Store number of operations currently in the queue."""
        return self._queue.qsize()

    @property
    def params(self):
        """Return PRUController parameters."""
        return self._parms

    @property
    def connected(self):
        """Store connection state."""
        return all((self.check_connected(id) for id in self._device_ids))

    def check_connected(self, device_id):
        """Return connection state of a device."""
        # TODO: may not be the true current connection state
        psupply = self._psupplies[device_id]
        return psupply.connected

    def timestamp_update(self):
        """Return tmestamp of last device update."""
        return self._timestamp_update

    # === queueing writes and local state copy reads ===

    # --- bsmp variables and parameters ---

    def read_variables(self, device_ids, variable_id=None):
        """
        Return device variables.

        Parameters
        ----------
        device_ids : int, tuple or list
            The BSMP device ids.
        variable_id : int or None, optional.
            The BSMP variable id selected. If not passed all device variables
            will be returned.

        Returns
        -------
        Selected BSMP device variable values.

        """
        # process device_ids
        if isinstance(device_ids, int):
            dev_ids = (device_ids, )
        else:
            dev_ids = device_ids

        # builds dict of requested values
        values = dict()
        for dev_id in dev_ids:
            if variable_id is None:
                values[dev_id] = self._psupplies[dev_id].variables
            else:
                values[dev_id] = \
                    self._psupplies[dev_id].get_variable(variable_id)

        # make copy
        with self._lock:
            if isinstance(device_ids, int):
                return _dcopy(values[device_ids])
            return _dcopy(values)

    def read_parameters(self, device_ids, parameter_id=None):
        """Return power supply parameters."""
        # process device_ids
        if isinstance(device_ids, int):
            dev_ids = (device_ids, )
        else:
            dev_ids = device_ids

        # builds dict of requested values
        values = dict()
        for dev_id in dev_ids:
            if parameter_id is None:
                values[dev_id] = self._psupplies[dev_id].parameters
            else:
                values[dev_id] = \
                    self._psupplies[dev_id].get_parameter(parameter_id)

        # make copy
        with self._lock:
            if isinstance(device_ids, int):
                return _dcopy(values[device_ids])
            return _dcopy(values)

    def exec_functions(self, device_ids, function_id, args=None):
        """Append BSMP function executions to opertations queue.

        Parameters
        ----------
        device_ids : int, tuple or list
            The BSMP device ids. It can be a list of ids or a singe id.
        function_id : int
            The BSMP function id to be executed for the devices.
        args : tuple, optional
            The list of BSMP function argument values

        Returns
        -------
        status : bool
            True is operation was queued or False, if operation was rejected
            because of the SOFBMode state.
        """
        # if in SOFBMode on, do not accept exec functions
        if self._sofb_mode:
            return False

        # prepare arguments
        if isinstance(device_ids, int):
            device_ids = (device_ids, )
        if args is None:
            args = (device_ids, function_id)
        else:
            args = (device_ids, function_id, args)

        # append bsmp function exec operation to queue
        operation = (self._bsmp_exec_function, args)
        self._queue.put(operation, block=False)
        return True

    def update_parameters(self, device_ids):
        """Update device parameters."""
        # if in SOFBMode on, do not accept comm. commands
        if self._sofb_mode:
            return False

        if isinstance(device_ids, int):
            device_ids = (device_ids, )

        # append function operation to queue
        args = (device_ids, )
        operation = (self._bsmp_read_parameter_values, args)
        self._queue.put(operation, block=False)
        return True

    # --- wfmref and scope curves ---

    def scope_update_auto_enable(self):
        """Enable wfmref and scope curves updates."""
        self._scope_update = True

    def scope_update_auto_disable(self):
        """Disable wfmref and scope curves updates."""
        self._scope_update = False

    @property
    def scope_update_auto(self):
        """Return state of scope_update_auto."""
        return self._scope_update

    def wfm_update(self, device_ids, interval=None):
        """Queue update wfm and scope curves."""
        if isinstance(device_ids, int):
            device_ids = (device_ids, )
        operation = (self._bsmp_update_wfm, (device_ids, interval, ))
        self._queue.put(operation, block=False)
        return True

    def wfmref_write(self, device_ids, data):
        """Write wfm curves."""
        # if in SOFBMode on, do not accept exec functions
        if self._sofb_mode:
            return False

        # prepare arguments
        if isinstance(device_ids, int):
            device_ids = (device_ids, )

        # append bsmp function exec operation to queue
        operation = (self._bsmp_wfmref_write, (device_ids, data))
        self._queue.put(operation, block=False)
        return True

    def wfmref_rb_read(self, device_id):
        """Return wfmref_rb curve."""
        psupply = self._psupplies[device_id]
        # NOTE: investigate whether lock and copy are really necessary!
        with self._lock:
            return _dcopy(psupply.wfmref_rb)

    def wfmref_read(self, device_id):
        """Return wfmref curve."""
        psupply = self._psupplies[device_id]
        with self._lock:
            curve = psupply.wfmref
            return _dcopy(curve)

    def wfmref_index(self, device_id):
        """Return current index into DSP selected curve."""
        psupply = self._psupplies[device_id]
        with self._lock:
            index = psupply.wfmref_index
        return index

    def scope_read(self, device_id):
        """Return Wfm-Mon curve."""
        psupply = self._psupplies[device_id]
        with self._lock:
            return _dcopy(psupply.scope)

    # --- SOFBCurrent parameters ---

    def sofb_mode_set(self, state):
        """Change SOFB mode: True or False."""
        self._sofb_mode = state
        if state:
            while not self._queue.empty():  # wait until queue is empty
                pass

    @property
    def sofb_mode(self):
        """Return SOFB mode."""
        return self._sofb_mode

    def sofb_current_set(self, value):
        """."""
        while not self._queue.empty():  # wait until queue is empty
            pass

        # execute SOFB setpoint
        self._bsmp_update_sofb_setpoint(value)

        return True

    @property
    def sofb_current_rb(self):
        """."""
        return self._udc.sofb_current_rb_get()

    @property
    def sofb_current_refmon(self):
        """."""
        return self._udc.sofb_current_refmon_get()

    @property
    def sofb_current_mon(self):
        """."""
        return self._udc.sofb_current_mon_get()

    def sofb_update_variables_state(self):
        """Update variables state mirror."""
        # do sofb update only if in SOFBMode On
        if not self._sofb_mode:
            return

        while not self._queue.empty():  # wait until queue is empty
            pass

        # select power supply dev_id for updating
        self._sofb_update_dev_idx = \
            (self._sofb_update_dev_idx + 1) % len(self._device_ids)
        dev_id = self._device_ids[self._sofb_update_dev_idx]

        # update variables state mirror for selected power supply
        self._bsmp_update_variables(dev_id)

    # --- scan and process loop methods ---

    def bsmp_scan(self):
        """Run scan one."""
        # select devices and variable group, defining the read group
        # operation to be performed
        operation = (self._bsmp_update, ())
        if self._queue.empty() or operation != self._queue.last_operation:
            self._queue.put(operation, block=False)
        else:
            # do not append if last operation is the same as last one
            # operation appended to queue
            pass

    def bsmp_init_communication(self):
        """."""
        # --- BSMP communication ---

        print()
        print('PRUController: bsmp initialization')
        # init time interval
        t0_ = _time()

        # reset power supply controllers (contains first BSMP comm)
        self._bsmp_reset_udc()

        # update state of PRUController from ps controller
        self._bsmp_init_update()

        # init thread structures
        self._init_threads()

        # time interval
        t1_ = _time()
        print('TIMING bsmp init [{:.3f} ms]\n'.format(
            1000*(t1_ - t0_)))

        # after all initializations, threads are started
        self._running = True
        if self._processing:
            self._queue.start()
        self._thread_scan.start()

    # --- private methods: initializations ---

    def _init_threads(self):

        fmt = '  - {:<20s} ({:^20s}) [{:09.3f}] ms'
        t0_ = _time()

        # define scan thread
        self._dev_idx_last_scanned = \
            len(self._device_ids)-1  # the next will be the first bsmp dev
        self._thread_scan = _Thread(target=self._loop_scan, daemon=True)

        dt_ = _time() - t0_
        print(fmt.format('init_threads', 'create structures', 1e3*dt_))

    @staticmethod
    def _init_udc(pru, psmodel_name, device_ids, freq):

        # create UDC
        udc = _UDC(pru, psmodel_name, device_ids)
        parms = udc.prucparms

        # create PSDevState objects
        psupplies = dict()
        for dev_id in device_ids:
            psupplies[dev_id] = _PSDevState(psbsmp=udc[dev_id])

        # bypass psmodel default frequencies
        if freq is not None:
            parms.FREQ_SCAN = freq

        # print info on scan frequency
        fstr = ('device_id:{:2d}, scan_freq: {:4.1f} Hz')
        for dev in device_ids:
            freq = 10.0 if freq is None else freq
            print(fstr.format(dev, parms.FREQ_SCAN))
        print()

        return udc, parms, psupplies

    def _init_check_version(self):
        if not self.connected:
            return
        for dev_id in self._device_ids:
            # V_FIRMWARE_VERSION should be defined for all BSMP devices
            psupply = self._psupplies[dev_id]
            _firmware_version_udc = psupply.get_variable(
                self._parms.CONST_PSBSMP.V_FIRMWARE_VERSION)
            _firmware_version_udc = \
                self._udc.parse_firmware_version(_firmware_version_udc)
            if 'Simulation' not in _firmware_version_udc and \
               _firmware_version_udc != _firmware_version_siriuspy:
                errmsg = (
                    'PRUController: Incompatible bsmp implementation version '
                    'for device id:{}')
                print(errmsg.format(dev_id))
                errmsg = 'lib version: {}'
                print(errmsg.format(_firmware_version_siriuspy))
                errmsg = 'udc version: {}'
                print(errmsg.format(_firmware_version_udc))
                print()
                # raise ValueError(errmsg)

    # --- private methods: scan and process ---

    def _loop_scan(self):
        while self._running:

            t0_ = _time()

            # run scan method once
            if self.scanning and \
               self._scan_interval != 0 and \
               not self._sofb_mode:
                self.bsmp_scan()

            # update scan interval
            self._scan_interval = self._get_scan_interval()

            # wait for time_interval
            dt_ = _time() - t0_
            if dt_ < self._scan_interval:
                _sleep(self._scan_interval - dt_)

            # update timestamp
            self._timestamp_update = _time()

    def _get_scan_interval(self):
        if self._parms.FREQ_SCAN == 0:
            return 0
        else:
            return 1.0/self._parms.FREQ_SCAN  # [s]

    def _serial_error(self, device_ids):
        # signal disconnected for device ids.
        for dev_id in device_ids:
            psupply = self._psupplies[dev_id]
            psupply._connected = False

    def _check_groups(self):
        group_ids = sorted(self._parms.groups.keys())

        # needs to have all default groups
        if len(group_ids) < 3:
            raise ValueError(('Invalid variable group definition: '
                              'default group missing!'))
        # needs to have all consecutive groups
        for i in range(len(group_ids)):
            if i not in group_ids:
                raise ValueError(('Invalid variable group definition: '
                                  'non-consecutive group ids!'))

    # --- private methods: BSMP UART communications ---

    def _bsmp_reset_udc(self):

        # set scan interval
        self._scan_interval = self._get_scan_interval()

        # initialize variable groups (first BSMP comm.)
        self._bsmp_init_devices()

    def _bsmp_init_devices(self):

        fmt = '  - {:<20s} ({:^20s}) [{:09.3f}] ms'

        # create vars groups list from dict
        groups = PRUController._dict2list_vargroups(self._parms.groups)

        # reset group of bsmp variables for all devices
        t0_ = _time()
        for psupply in self._psupplies.values():
            psupply.reset_variables_groups(groups)
        dt_ = _time() - t0_
        print(fmt.format('bsmp_init_devices', 'reset groups', 1e3*dt_))

        # update psupply groups
        t0_ = _time()
        for psupply in self._psupplies.values():
            psupply.update_groups(interval=0.0)
        dt_ = _time() - t0_
        print(fmt.format('bsmp_init_devices', 'update groups', 1e3*dt_))

        # disable DSP from writting to bufsample (uses first device)
        t0_ = _time()
        self._udc.bufsample_disable()
        dt_ = _time() - t0_
        print(fmt.format('bsmp_init_devices', 'bufsample_disable', 1e3*dt_))

    def _bsmp_update(self):

        try:
            # update variables
            self._bsmp_update_variables()

            # update device wfm curves cyclically
            if self._scope_update:
                self._scope_update_dev_idx = \
                    (self._scope_update_dev_idx + 1) % len(self._device_ids)
                dev_id = self._device_ids[self._scope_update_dev_idx]
                self._bsmp_update_wfm(dev_id)

        except _socket_timeout:
            print('!!! {} : socket timeout !!!'.format(_get_timestamp()))

    def _bsmp_update_variables(self, dev_id=None):
        if dev_id is None:
            psupplies = self._psupplies.values()
        else:
            psupplies = (self._psupplies[dev_id], )

        for psupply in psupplies:
            try:
                t0_ = _time()
                psupply.update_variables(interval=0.0)
            except _SerialError as err:
                # no serial connection !
                dt_ = _time() - t0_
                print(
                    f'!!! {_get_timestamp()}: {err}. '
                    f'it took {dt_*1000:.3f} ms in bsmp_update_variables.'
                )

    def _bsmp_update_wfm(self, device_id):
        """Read curve from devices."""
        psupplies = self._psupplies

        try:
            t0_ = _time()
            psupply = psupplies[device_id]
            psupply.update_wfm()
        except _SerialError as err:
            # no serial connection !
            dt_ = _time() - t0_
            print(
                f'!!! {_get_timestamp()}: {err}. '
                f'it took {dt_*1000:.3f} ms in bsmp_update_wfm.'
            )

        # stores updated psupplies dict
        self._psupplies = psupplies  # atomic operation

    def _bsmp_update_sofb_setpoint(self, value):

        # execute sofb current setpoint
        self._udc.sofb_current_set(value)

        # update sofb state
        self._udc.sofb_update()

        # # update all other device parameters
        # self._bsmp_update()

        # print('{:<30s} : {:>9.3f} ms'.format(
        #     'PRUC._bsmp_update_sofb_setpoint (end)', 1e3*(_time() % 1)))

    def _bsmp_wfmref_write(self, device_ids, curve):
        """Write wfmref curve to devices."""
        try:
            # write curves
            for dev_id in device_ids:
                psupply = self._psupplies[dev_id]
                # sends new wfmref curve to power supply then reads back
                # curve in UDC memory
                curve = psupply.psbsmp.wfmref_write(curve)
                with self._lock:
                    psupply.wfmref_rb = curve
                    psupply.update_wfm(interval=0.0)
        except (_SerialError, IndexError):
            print('PRUController: bsmp_wfmref_write error!')
            self._serial_error(device_ids)

    def _bsmp_exec_function(self, device_ids, function_id, args=None):
        # --- send func exec request to serial line

        # BSMP device's 'execute_function' needs to lock code execution
        # so as to avoid more than one thread reading each other's responses.
        # class BSMP method 'request' should always do that

        ack, data = dict(), dict()

        # --- send requests to serial line
        try:
            for dev_id in device_ids:
                resp = self._udc[dev_id].execute_function(function_id, args)
                ack[dev_id], data[dev_id] = resp
                # check anomalous response
                if ack[dev_id] != _const_bsmp.ACK_OK:
                    print('PRUController: anomalous response !')
                    datum = data[dev_id]
                    if isinstance(datum, str):
                        datum = ord(datum)
                    self._udc[dev_id].anomalous_response(
                        _const_bsmp.CMD_EXECUTE_FUNCTION, ack[dev_id],
                        device_id=dev_id, function_id=function_id,
                        data_type=type(datum), data=datum)
        except _SerialError:
            return None
        except TypeError:
            print('--- PRUController debug ---')
            print('device_ids  : ', device_ids)
            print('dev_id      : ', dev_id)
            print('function_id : ', function_id)
            print('resp        : ', resp)
            print('data        : ', data[dev_id])
            raise

        # --- check if all function executions succeeded.
        success = True
        for dev_id in device_ids:
            connected = (ack[dev_id] == _const_bsmp.ACK_OK)
            success &= connected

        if success:
            return data
        else:
            return None

    def _bsmp_init_update(self):

        fmt = '  - {:<20s} ({:^20s}) [{:09.3f}] ms'

        # initialize variables_values, a mirror state of BSMP devices
        t0_ = _time()
        self._bsmp_init_variable_values()
        dt_ = _time() - t0_
        print(fmt.format('bsmp_init_update', 'variable_values', 1e3*dt_))

        # initialize ps curves
        t0_ = _time()
        self._bsmp_init_wfm()
        dt_ = _time() - t0_
        print(fmt.format('bsmp_init_update', 'waveform_values', 1e3*dt_))

        # initialize sofb
        t0_ = _time()
        self._bsmp_init_sofb_values()
        dt_ = _time() - t0_
        print(fmt.format('bsmp_init_update', 'sofb_values', 1e3*dt_))

        # initialize parameters
        t0_ = _time()
        self._bsmp_init_parameter_values()
        dt_ = _time() - t0_
        print(fmt.format('bsmp_init_update', 'parameter_values', 1e3*dt_))

    def _bsmp_init_wfm(self):

        for psupply in self._psupplies.values():

            # update
            psupply.update_wfm(interval=0.0)

            # raise ValueError('!!! Debug stop !!!')

            # registers RB using psupply object
            psupply.wfmref_rb = psupply.wfmref

            # raise ValueError('!!! Debug stop !!!')

        # raise ValueError('!!! Debug stop !!!')

        # enable bufsample
        # self._udc.bufsample_enable()

        # raise ValueError('!!! Debug stop !!!')

    def _bsmp_init_variable_values(self):

        # init psupplies variables
        for psupply in self._psupplies.values():
            psupply.update_variables(interval=0.0)

    def _bsmp_init_parameter_values(self):
        # init psupplies parameters
        self._bsmp_read_parameter_values()

    def _bsmp_init_sofb_values(self):

        self._udc.sofb_update()

    def _bsmp_read_parameter_values(self, device_ids=None):
        device_ids = device_ids or self._device_ids
        for dev_id in device_ids:
            psupply = self._psupplies[dev_id]
            # read psupplies parameters
            psupply.update_parameters(interval=0.0)

    @staticmethod
    def _dict2list_vargroups(groups_dict):
        group_ids = sorted(groups_dict.keys())
        if len(group_ids) < 3:  # needs to have all default groups
            print('PRUController: Incorrect variables group definition: '
                  'it does not have all three standard groups!')
            raise ValueError
        for i in range(len(group_ids)):  # consecutive?
            if i not in group_ids:
                print('PRUController: Incorrect variables group definition: '
                      'it does not have consecutive group ids!')
                raise ValueError
        # create list of variable ids
        groups_list = [groups_dict[gid] for gid in group_ids]
        return groups_list



"""Define DCCT device high level PVs database.

This is not a primary source database. Primary sources can be found in:
 - Digital Multimeter: https://github.com/lnls-dig/dmm7510-epics-ioc
"""

from ... import csdev as _csdev


# --- Enumeration Types ---

class ETypes(_csdev.ETypes):
    """Local enumerate types."""

    STOREDEBEAM = ('False', 'True')
    MEASMODE_SEL = ('Normal', 'Fast')
    MEASMODE_STS = ('Unknown', 'Normal', 'Fast')
    MEASTRG_SEL = ('None', 'External', 'InLevel', 'Software')
    MEASTRG_STS = ('Unknown', 'External', 'InLevel', 'Software', 'None')
    IMPED = ('AUTO', '10MOhm')
    AVGFILTERTYP = ('Repeat', 'Moving')
    RANGE = ('20 A', '2 A', '200 mA', '20 mA')


_et = ETypes  # syntactic sugar


# --- Const class ---

class Const(_csdev.Const):
    """Const class defining DCCTs constants and Enum types."""

    MeasModeSel = _csdev.Const.register('MeasModeSel', _et.MEASMODE_SEL)
    MeasModeSts = _csdev.Const.register('MeasModeSts', _et.MEASMODE_STS)
    AvgFilterTyp = _csdev.Const.register('AvgFilterTyp', _et.AVGFILTERTYP)


_c = Const  # syntactic sugar


# --- Database ---

def get_dcct_database():
    """Return DCCT device database."""
    pvs_database = {
        # Average Current Measurement
        'Current-Mon': {'type': 'float', 'unit': 'mA', 'prec': 6},
        'Timestamp-Mon': {'type': 'float', 'unit': 's', 'prec': 6},
        'CurrHstr-Mon': {'type': 'float', 'count': 1000, 'unit': 'mA'},
        'TimestampHstr-Mon': {'type': 'float', 'count': 1000, 'unit': 's'},
        'ClearHstr-Cmd': {'type': 'int', 'value': 0},
        'RawReadings-Mon': {'type': 'float', 'count': 1000000, 'unit': 'mA'},
        'StoredEBeam-Mon': {'type': 'enum', 'enums': _et.STOREDEBEAM},

        # Measurement Reliability
        'ReliableMeas-Mon': {'type': 'int',
                             'hihi': 15, 'high': 1, 'hilim': 15,
                             'lolim': 0, 'low': 0, 'lolo': 0},
        'ReliableMeasLabels-Cte': {'type': 'string', 'count': 4},

        # Measurement Trigger Source and Timer Configuration
        'Enbl-Sel': {'type': 'enum', 'enums': _et.OFF_ON},
        'Enbl-Sts': {'type': 'enum', 'enums': _et.OFF_ON},
        'Abort-Cmd': {'type': 'int', 'value': 0},
        'MeasTrg-Sel': {'type': 'enum', 'enums': _et.MEASTRG_SEL},
        'MeasTrg-Sts': {'type': 'enum', 'enums': _et.MEASTRG_STS},

        # Measurement Mode Selection
        'MeasMode-Sel': {'type': 'enum', 'enums': _et.MEASMODE_SEL},
        'MeasMode-Sts': {'type': 'enum', 'enums': _et.MEASMODE_STS},

        # Measurement and Filter Settings | Normal Mode
        'SampleCnt-SP': {'type': 'float', 'hihi': 5000, 'high': 5000,
                         'hilim': 5000, 'lolim': 1, 'low': 1, 'lolo': 1},
        'SampleCnt-RB': {'type': 'float', 'hihi': 5000, 'high': 5000,
                         'hilim': 5000, 'lolim': 1, 'low': 1, 'lolo': 1},
        'MeasPeriod-SP': {'type': 'float', 'unit': 's', 'prec': 7,
                          'value': 0.01667, 'hihi': 10, 'high': 10,
                          'hilim': 10, 'lolim': 0.0000084,
                          'low': 0.0000084, 'lolo': 0.0000084},
        'MeasPeriod-RB': {'type': 'float', 'unit': 's', 'prec': 7,
                          'value': 0.01667, 'hihi': 10, 'high': 10,
                          'hilim': 10, 'lolim': 0.0000084,
                          'low': 0.0000084, 'lolo': 0.0000084},
        'MeasUpdatePeriod-Mon': {'type': 'float', 'unit': 's', 'prec': 7},
        'Imped-Sel': {'type': 'enum', 'enums': _et.IMPED, 'value': 0},
        'Imped-Sts': {'type': 'enum', 'enums': _et.IMPED, 'value': 0},
        'LineSync-Sel': {'type': 'enum', 'enums': _et.OFF_ON, 'value': 0},
        'LineSync-Sts': {'type': 'enum', 'enums': _et.OFF_ON, 'value': 0},
        'RelEnbl-Sel': {'type': 'enum', 'enums': _et.OFF_ON, 'value': 0},
        'RelEnbl-Sts': {'type': 'enum', 'enums': _et.OFF_ON, 'value': 0},
        'RelLvl-SP': {'type': 'float', 'prec': 16,
                      'high': 1000000, 'hihi': 1000000,
                      'hilim': 1000000, 'lolim': -1000000,
                      'low': -1000000, 'lolo': -1000000},
        'RelLvl-RB': {'type': 'float', 'prec': 16,
                      'high': 1000000, 'hihi': 1000000,
                      'hilim': 1000000, 'lolim': -1000000,
                      'low': -1000000, 'lolo': -1000000},
        'RelAcq-Cmd': {'type': 'int', 'value': 0},
        'AvgFilterEnbl-Sel': {'type': 'enum', 'enums': _et.OFF_ON},
        'AvgFilterEnbl-Sts': {'type': 'enum', 'enums': _et.OFF_ON},
        'AvgFilterCnt-SP': {'type': 'float', 'hihi': 100, 'high': 100,
                            'hilim': 100, 'lolim': 1, 'low': 1, 'lolo': 1},
        'AvgFilterCnt-RB': {'type': 'float', 'hihi': 100, 'high': 100,
                            'hilim': 100, 'lolim': 1, 'low': 1, 'lolo': 1},
        'AvgFilterTyp-Sel': {'type': 'enum', 'enums': _et.AVGFILTERTYP},
        'AvgFilterTyp-Sts': {'type': 'enum', 'enums': _et.AVGFILTERTYP},
        'AvgFilterWind-SP': {'type': 'float', 'unit': '%',
                             'hihi': 10, 'high': 10, 'hilim': 10,
                             'lolim': 0, 'low': 0, 'lolo': 0},
        'AvgFilterWind-RB': {'type': 'float', 'unit': '%',
                             'hihi': 10, 'high': 10, 'hilim': 10,
                             'lolim': 0, 'low': 0, 'lolo': 0},

        # Measurement Settings | Fast Mode
        'FastSampleCnt-SP': {'type': 'float', 'hihi': 5000, 'high': 5000,
                             'hilim': 5000, 'lolim': 1, 'low': 1, 'lolo': 1},
        'FastSampleCnt-RB': {'type': 'float', 'hihi': 5000, 'high': 5000,
                             'hilim': 5000, 'lolim': 1, 'low': 1, 'lolo': 1},
        'FastMeasPeriod-SP': {'type': 'float', 'unit': 's', 'prec': 7,
                              'value': 0.001, 'hihi': 5, 'high': 5,
                              'hilim': 5, 'lolim': 0.000001,
                              'low': 0.000001, 'lolo': 0.000001},
        'FastMeasPeriod-RB': {'type': 'float', 'unit': 's', 'prec': 7,
                              'value': 0.001, 'hihi': 5, 'high': 5,
                              'hilim': 5, 'lolim': 0.000001,
                              'low': 0.000001, 'lolo': 0.000001},
        'FastImped-Sel': {'type': 'enum', 'enums': _et.IMPED, 'value': 0},
        'FastImped-Sts': {'type': 'enum', 'enums': _et.IMPED, 'value': 0},
        'FastRelEnbl-Sel': {'type': 'enum', 'enums': _et.OFF_ON, 'value': 0},
        'FastRelEnbl-Sts': {'type': 'enum', 'enums': _et.OFF_ON, 'value': 0},
        'FastRelLvl-SP': {'type': 'float', 'prec': 16,
                          'high': 1000000, 'hihi': 1000000,
                          'hilim': 1000000, 'lolim': -1000000,
                          'low': -1000000, 'lolo': -1000000},
        'FastRelLvl-RB': {'type': 'float', 'prec': 16,
                          'high': 1000000, 'hihi': 1000000,
                          'hilim': 1000000, 'lolim': -1000000,
                          'low': -1000000, 'lolo': -1000000},
        'FastRelAcq-Cmd': {'type': 'int', 'value': 0},

        # Current Measurement Range and Threshold Detection | Both Modes
        'Range-Sel': {'type': 'enum', 'enums': _et.RANGE},
        'Range-Sts': {'type': 'enum', 'enums': _et.RANGE},
        'LowLimEnbl-Sel': {'type': 'enum', 'enums': _et.OFF_ON},
        'LowLimEnbl-Sts': {'type': 'enum', 'enums': _et.OFF_ON},
        'CurrThold-SP': {'type': 'float', 'unit': 'mA', 'prec': 6,
                         'hihi': 2000, 'high': 2000, 'hilim': 2000,
                         'lolim': -2000, 'low': -2000, 'lolo': -2000},
        'CurrThold-RB': {'type': 'float', 'unit': 'mA', 'prec': 6,
                         'hihi': 2000, 'high': 2000, 'hilim': 2000,
                         'lolim': -2000, 'low': -2000, 'lolo': -2000},
        'HFReject-Sel': {'type': 'enum', 'enums': _et.OFF_ON},
        'HFReject-Sts': {'type': 'enum', 'enums': _et.OFF_ON},

        # DCCT Calibration Setting
        'Test-Sel': {'type': 'enum', 'enums': _et.OFF_ON},
        'Test-Sts': {'type': 'enum', 'enums': _et.OFF_ON},

        # Multimeter Setup
        'Download-Cmd': {'type': 'int', 'value': 0}
    }
    return pvs_database



"""Module with BO ramp and SI mig classes."""

from copy import deepcopy as _dcopy
import numpy as _np

from ..pwrsupply.csdev import MAX_WFMSIZE_FBP as _MAX_WFMSIZE_FBP
from ..pwrsupply.csdev import MAX_WFMSIZE_OTHERS as \
    _MAX_WFMSIZE_OTHERS

from ..search import PSSearch as _PSSearch
from ..namesys import SiriusPVName
from ..clientconfigdb import ConfigDBDocument as _ConfigDBDocument
from ..magnet.util import \
    get_magnet_family_name as _get_magnet_family_name

from .exceptions import RampError as _RampError, \
    RampInvalidDipoleWfmParms as _RampInvalidDipoleWfmParms, \
    RampInvalidNormConfig as _RampInvalidNormConfig, \
    RampInvalidRFParms as _RampInvalidRFParms

from .util import MAX_RF_RAMP_DURATION as _MAX_RF_RAMP_DURATION, \
    NOMINAL_STRENGTHS as _NOMINAL_STRENGTHS
from .waveform import WaveformDipole as _WaveformDipole, \
    Waveform as _Waveform


class BoosterNormalized(_ConfigDBDocument):
    """Booster normalized configuration."""

    def __init__(self, name=None):
        """Constructor."""
        super().__init__('bo_normalized', name=name)
        self._value = self.get_value_from_template()
        self._orig_value = None

        self._psnames2index = dict()
        for index, data in enumerate(self._value['pvs']):
            psname = SiriusPVName(data[0]).device_name
            self._psnames2index[psname] = index

    @property
    def psnames(self):
        """."""
        return list(self._psnames2index.keys())

    def load(self):
        """."""
        super().load()
        self._orig_value = self.value

    def save(self, **kws):
        """."""
        super().save(**kws)
        self._orig_value = self.value

    def verify_synchronized(self, psname=None):
        """."""
        if not self._orig_value:
            return
        modified = False
        for i in range(len(self._value['pvs'])):
            if self._value['pvs'][i][1] != self._orig_value['pvs'][i][1]:
                modified = True
                break
        self._synchronized = not modified

    def _get_item(self, psname):
        index = self._psnames2index[psname]
        value = self._value['pvs'][index][1]
        return float(value)

    def _set_item(self, psname, value):
        """Set configuration item."""
        index = self._psnames2index[psname]
        if self._value['pvs'][index][1] != value:
            self._value['pvs'][index][1] = value
            self.verify_synchronized()

    def _set_value(self, value):
        """Set configuration."""
        if isinstance(value, dict):
            for key, val in value.items():
                if key == 'label':
                    self.name = val
                else:
                    self[key] = val
        else:
            super()._set_value(value)
        self.verify_synchronized()

    def __str__(self):
        """Return string representation of configuration."""
        if not self._value:
            rst = 'name: {}'.format(self.name)
            return rst
        rst = ''
        k = [data[0] for data in self._value['pvs']]
        val1 = [data[1] for data in self._value['pvs']]
        val2 = [data[2] for data in self._value['pvs']]
        maxlen = max(tuple(len(ky) for ky in k) + (len('name'),))
        fmtstr1 = '{:<'+str(maxlen)+'}, {:+.6f}, {:+.6f}\n'
        fmtstr2 = '{:<'+str(maxlen)+'}: {}\n'
        rst = fmtstr2.format('name', self.name)
        for i, kval in enumerate(k):
            rst += fmtstr1.format(kval, val1[i], val2[i])
        return rst

    def __eq__(self, value):
        for psname in self.psnames:
            if psname in ['BO-Fam:PS-B-1', 'BO-Fam:PS-B-2']:
                continue
            if not _np.isclose(self[psname], value[psname], atol=1e-5):
                return False
        return True


class BoosterRamp(_ConfigDBDocument):
    """Booster ramp class."""

    PSNAMES = _PSSearch.get_psnames({'sec': 'BO', 'dis': 'PS'})
    PSNAME_DIPOLES = ('BO-Fam:PS-B-1', 'BO-Fam:PS-B-2')
    PSNAME_DIPOLE_REF = PSNAME_DIPOLES[0]

    def __init__(self, name=None, auto_update=False):
        """Constructor."""
        super().__init__('bo_ramp', name=name)
        self._auto_update = auto_update
        self._ps_waveforms = dict()
        self._nominal_strengths = dict()
        self._orig_value = None
        self.value = self.get_value_from_template()

    # --- ConfigSrv API ---

    def load(self):
        """Load configuration from config server."""
        # load booster ramp configuration
        super().load()
        self._orig_value = self.value
        self._invalidate_ps_waveforms(True)

    def save(self, new_name=None):
        """."""
        super().save(new_name=new_name)
        self._orig_value = self.value

    # ---- ps_normalized_configs ----

    @property
    def ps_normalized_configs(self):
        """List of ps normalized config."""
        return _dcopy(self._value['ps_normalized_configs*'])

    def ps_normalized_configs_set(self, value):
        """Receive a dict of times as keys and ps2strengths dict as values."""
        self._value['ps_normalized_configs*'] = _dcopy(value)
        self.verify_synchronized()
        self._invalidate_ps_waveforms()

    @property
    def ps_normalized_configs_times(self):
        """Return time instants corresponding to ps normalized configs."""
        return [float(k) for k in self._value['ps_normalized_configs*'].keys()]

    @property
    def ps_normalized_configs_labels(self):
        """Return names corresponding to ps normalized configs."""
        return [n['label'] for n in
                self._value['ps_normalized_configs*'].values()]

    def ps_normalized_configs_delete(self, time):
        """Delete a ps normalized config by its time."""
        self._value['ps_normalized_configs*'].pop('{:.3f}'.format(time))
        self.verify_synchronized()
        self._invalidate_ps_waveforms()

    def ps_normalized_configs_insert(self, time, label='',
                                     psname2strength=dict()):
        """Insert a ps normalized configuration."""
        # verify if there is no config in this time
        otimes = sorted(self.ps_normalized_configs_times)
        if time in otimes:
            raise _RampInvalidNormConfig(
                'There is already a configuration at this time.')

        # create nconfig data
        nconfig = dict(label=label)
        # interpolate strengths, if necessary
        if not psname2strength:
            psname2strength = dict()
            if otimes:
                for psname in self.PSNAMES:
                    if psname in self.PSNAME_DIPOLES:
                        continue
                    ovalues = list()
                    for _time in otimes:
                        str_time = '{:.3f}'.format(_time)
                        ncv = self._value['ps_normalized_configs*'][str_time]
                        ovalues.append(ncv[psname])
                    psname2strength[psname] = _np.interp(time, otimes, ovalues)
            else:
                psname2strength = self.ps_normalized_config_nominal_values
        nconfig.update(psname2strength)
        nconfig = self._update_ps_normalized_config_energy(nconfig, time)
        self._value['ps_normalized_configs*']['{:.3f}'.format(time)] = nconfig

        self.verify_synchronized()
        self._invalidate_ps_waveforms()

    def ps_normalized_configs_change_time(self, time, new_time,
                                          change_energy=False):
        """Change the time of an existing config."""
        otimes = self.ps_normalized_configs_times
        if new_time in otimes:
            raise _RampInvalidNormConfig(
                'There is already a configuration at this time.')
        str_time = '{:.3f}'.format(time)
        str_new_time = '{:.3f}'.format(new_time)

        # set config energy appropriately if needed
        if change_energy:
            self._update_ps_normalized_config_energy(
                self._value['ps_normalized_configs*'][str_time], new_time)

        nconfig = self._value['ps_normalized_configs*'].pop(str_time)
        self._value['ps_normalized_configs*'][str_new_time] = nconfig
        self.verify_synchronized()
        self._invalidate_ps_waveforms()

    def ps_normalized_configs_change_label(self, time, new_label):
        """Change the label of an existing config given its time."""
        str_time = '{:.3f}'.format(time)
        self._value['ps_normalized_configs*'][str_time]['label'] = new_label
        self.verify_synchronized()

    @property
    def ps_normalized_config_nominal_values(self):
        """."""
        if not self._nominal_strengths:
            psname2strength = dict()
            for psname in self.PSNAMES:
                if psname in _NOMINAL_STRENGTHS:
                    psname2strength[psname] = _NOMINAL_STRENGTHS[psname]
                else:
                    psname2strength[psname] = 0.0
            self._nominal_strengths = psname2strength
        return self._nominal_strengths

    # ---- ps ramp parameters ----

    @property
    def ps_ramp_duration(self):
        """Power supplies ramp duration."""
        return self._value['ps_ramp']['duration']

    @ps_ramp_duration.setter
    def ps_ramp_duration(self, value):
        """Set power supplies duration."""
        value = float(value)
        if value != self._value['ps_ramp']['duration']:
            if not self.ps_ramp_rampdown_stop_time < value:
                raise _RampInvalidDipoleWfmParms(
                    'Invalid duration for waveforms.')
            self._value['ps_ramp']['duration'] = value
            self.verify_synchronized()
            self._invalidate_ps_waveforms(True)

    @property
    def ps_ramp_wfm_nrpoints_fams(self):
        """Power supplies waveform number of points."""
        rdip = self._value['ps_ramp']
        return rdip['wfm_nrpoints_fams']

    @ps_ramp_wfm_nrpoints_fams.setter
    def ps_ramp_wfm_nrpoints_fams(self, value):
        """Set power supplies waveform number of points for families."""
        value = int(value)
        rdip = self._value['ps_ramp']
        if value != rdip['wfm_nrpoints_fams']:
            if not 1 <= value <= _MAX_WFMSIZE_OTHERS:
                raise _RampInvalidDipoleWfmParms(
                    'Invalid number of points for waveforms.')
            rdip['wfm_nrpoints_fams'] = value
            self.verify_synchronized()
            self._invalidate_ps_waveforms(True)

    @property
    def ps_ramp_wfm_nrpoints_corrs(self):
        """Power supplies waveform number of points for correctors."""
        rdip = self._value['ps_ramp']
        return rdip['wfm_nrpoints_corrs']

    @ps_ramp_wfm_nrpoints_corrs.setter
    def ps_ramp_wfm_nrpoints_corrs(self, value):
        """Set power supplies waveform number of points for correctors."""
        value = int(value)
        rdip = self._value['ps_ramp']
        if value != rdip['wfm_nrpoints_corrs']:
            if not 1 <= value <= _MAX_WFMSIZE_FBP:
                raise _RampInvalidDipoleWfmParms(
                    'Invalid number of points for waveforms.')
            rdip['wfm_nrpoints_corrs'] = value
            self.verify_synchronized()
            self._invalidate_ps_waveforms(True)

    @property
    def ps_ramp_times(self):
        """Return ps ramp times."""
        vals = (self.ps_ramp_rampup1_start_time,
                self.ps_ramp_rampup2_start_time,
                self.ps_ramp_rampdown_start_time,
                self.ps_ramp_rampdown_stop_time)
        return vals

    @property
    def ps_ramp_energies(self):
        """Return ps ramp times."""
        vals = (self.ps_ramp_rampup1_start_energy,
                self.ps_ramp_rampup2_start_energy,
                self.ps_ramp_rampdown_start_energy,
                self.ps_ramp_rampdown_stop_energy)
        return vals

    @property
    def ps_ramp_start_energy(self):
        """Return."""
        return self._value['ps_ramp']['start_energy']

    @ps_ramp_start_energy.setter
    def ps_ramp_start_energy(self, value):
        """Return."""
        value = float(value)
        rdip = self._value['ps_ramp']
        if value != rdip['start_energy']:
            wfm = self._create_new_ps_waveform_dipole()
            wfm.start_energy = value
            rdip['start_energy'] = wfm.start_energy
            self.verify_synchronized()
            self._invalidate_ps_waveforms(True)

    @property
    def ps_ramp_rampup1_start_energy(self):
        """Return."""
        return self._value['ps_ramp']['rampup1_start_energy']

    @ps_ramp_rampup1_start_energy.setter
    def ps_ramp_rampup1_start_energy(self, value):
        """Return."""
        value = float(value)
        rdip = self._value['ps_ramp']
        if value != rdip['rampup1_start_energy']:
            wfm = self._create_new_ps_waveform_dipole()
            wfm.rampup1_start_energy = value
            rdip['rampup1_start_energy'] = wfm.rampup1_start_energy
            self.verify_synchronized()
            self._invalidate_ps_waveforms(True)

    @property
    def ps_ramp_rampup1_start_time(self):
        """Return."""
        return self._value['ps_ramp']['rampup1_start_time']

    @ps_ramp_rampup1_start_time.setter
    def ps_ramp_rampup1_start_time(self, value):
        """Return."""
        self._check_valid_time(value)
        rdip = self._value['ps_ramp']
        if value != rdip['rampup1_start_time']:
            rdip['rampup1_start_time'] = value
            self.verify_synchronized()
            self._invalidate_ps_waveforms(True)

    @property
    def ps_ramp_rampup1_slope(self):
        """Return rampup1 slope."""
        self._update_ps_waveform(self.PSNAME_DIPOLE_REF)
        return self._ps_waveforms[self.PSNAME_DIPOLE_REF].rampup1_slope

    @property
    def ps_ramp_rampup2_start_energy(self):
        """Return."""
        return self._value['ps_ramp']['rampup2_start_energy']

    @ps_ramp_rampup2_start_energy.setter
    def ps_ramp_rampup2_start_energy(self, value):
        """Return."""
        value = float(value)
        rdip = self._value['ps_ramp']
        if value != rdip['rampup2_start_energy']:
            wfm = self._create_new_ps_waveform_dipole()
            wfm.rampup2_start_energy = value
            rdip['rampup2_start_energy'] = wfm.rampup2_start_energy
            self.verify_synchronized()
            self._invalidate_ps_waveforms(True)

    @property
    def ps_ramp_rampup2_start_time(self):
        """Return."""
        return self._value['ps_ramp']['rampup2_start_time']

    @ps_ramp_rampup2_start_time.setter
    def ps_ramp_rampup2_start_time(self, value):
        """Return."""
        self._check_valid_time(value)
        rdip = self._value['ps_ramp']
        if value != rdip['rampup2_start_time']:
            rdip['rampup2_start_time'] = value
            self.verify_synchronized()
            self._invalidate_ps_waveforms(True)

    @property
    def ps_ramp_rampup2_slope(self):
        """Return rampup2 slope."""
        self._update_ps_waveform(self.PSNAME_DIPOLE_REF)
        return self._ps_waveforms[self.PSNAME_DIPOLE_REF].rampup2_slope

    @property
    def ps_ramp_rampup_smooth_energy(self):
        """Return."""
        return self._value['ps_ramp']['rampup_smooth_energy']

    @ps_ramp_rampup_smooth_energy.setter
    def ps_ramp_rampup_smooth_energy(self, value):
        """Return."""
        rdip = self._value['ps_ramp']
        if value != rdip['rampup_smooth_energy']:
            rdip['rampup_smooth_energy'] = value
            self.verify_synchronized()
            self._invalidate_ps_waveforms(True)

    @property
    def ps_ramp_rampup_smooth_intvl(self):
        """Return."""
        return self._value['ps_ramp']['rampup_smooth_intvl']

    @ps_ramp_rampup_smooth_intvl.setter
    def ps_ramp_rampup_smooth_intvl(self, value):
        """Return."""
        rdip = self._value['ps_ramp']
        if value != rdip['rampup_smooth_intvl']:
            rdip['rampup_smooth_intvl'] = value
            self.verify_synchronized()
            self._invalidate_ps_waveforms(True)

    @property
    def ps_ramp_rampdown_start_energy(self):
        """Return."""
        return self._value['ps_ramp']['rampdown_start_energy']

    @ps_ramp_rampdown_start_energy.setter
    def ps_ramp_rampdown_start_energy(self, value):
        """Return."""
        value = float(value)
        rdip = self._value['ps_ramp']
        if value != rdip['rampdown_start_energy']:
            wfm = self._create_new_ps_waveform_dipole()
            wfm.rampdown_start_energy = value
            rdip['rampdown_start_energy'] = wfm.rampdown_start_energy
            self.verify_synchronized()
            self._invalidate_ps_waveforms(True)

    @property
    def ps_ramp_rampdown_start_time(self):
        """Return."""
        return self._value['ps_ramp']['rampdown_start_time']

    @ps_ramp_rampdown_start_time.setter
    def ps_ramp_rampdown_start_time(self, value):
        """Return."""
        self._check_valid_time(value)
        rdip = self._value['ps_ramp']
        if value != rdip['rampdown_start_time']:
            rdip['rampdown_start_time'] = value
            self.verify_synchronized()
            self._invalidate_ps_waveforms(True)

    @property
    def ps_ramp_rampdown_stop_energy(self):
        """Return."""
        return self._value['ps_ramp']['rampdown_stop_energy']

    @ps_ramp_rampdown_stop_energy.setter
    def ps_ramp_rampdown_stop_energy(self, value):
        """Return."""
        value = float(value)
        rdip = self._value['ps_ramp']
        if value != rdip['rampdown_stop_energy']:
            wfm = self._create_new_ps_waveform_dipole()
            wfm.rampdown_stop_energy = value
            rdip['rampdown_stop_energy'] = wfm.rampdown_stop_energy
            self.verify_synchronized()
            self._invalidate_ps_waveforms(True)

    @property
    def ps_ramp_rampdown_stop_time(self):
        """Return."""
        return self._value['ps_ramp']['rampdown_stop_time']

    @ps_ramp_rampdown_stop_time.setter
    def ps_ramp_rampdown_stop_time(self, value):
        """Return."""
        self._check_valid_time(value)
        rdip = self._value['ps_ramp']
        if value != rdip['rampdown_stop_time']:
            rdip['rampdown_stop_time'] = value
            self.verify_synchronized()
            self._invalidate_ps_waveforms(True)

    @property
    def ps_ramp_rampdown_slope(self):
        """Return rampup2 slope."""
        self._update_ps_waveform(self.PSNAME_DIPOLE_REF)
        return self._ps_waveforms[self.PSNAME_DIPOLE_REF].rampdown_slope

    @property
    def ps_ramp_rampdown_smooth_energy(self):
        """Return."""
        return self._value['ps_ramp']['rampdown_smooth_energy']

    @ps_ramp_rampdown_smooth_energy.setter
    def ps_ramp_rampdown_smooth_energy(self, value):
        """Return."""
        rdip = self._value['ps_ramp']
        if value != rdip['rampdown_smooth_energy']:
            rdip['rampdown_smooth_energy'] = value
            self.verify_synchronized()
            self._invalidate_ps_waveforms(True)

    @property
    def ps_ramp_rampdown_smooth_intvl(self):
        """Return."""
        return self._value['ps_ramp']['rampdown_smooth_intvl']

    @ps_ramp_rampdown_smooth_intvl.setter
    def ps_ramp_rampdown_smooth_intvl(self, value):
        """Return."""
        rdip = self._value['ps_ramp']
        if value != rdip['rampdown_smooth_intvl']:
            rdip['rampdown_smooth_intvl'] = value
            self.verify_synchronized()
            self._invalidate_ps_waveforms(True)

    # ---- rf ramp parameters ----

    @property
    def rf_ramp_duration(self):
        """RF ramp duration."""
        d = self.rf_ramp_bottom_duration + \
            self.rf_ramp_rampup_duration + \
            self.rf_ramp_top_duration + \
            self.rf_ramp_rampdown_duration
        return d

    @property
    def rf_ramp_times(self):
        """Time instants to define RF ramp."""
        time = (
            self.ti_params_rf_ramp_delay,      # = RF ramp start, bottom start
            self.rf_ramp_rampup_start_time,    # = RF bottom stop
            self.rf_ramp_rampup_stop_time,     # = RF top start
            self.rf_ramp_rampdown_start_time,  # = RF top stop
            self.rf_ramp_duration)             # = RF rampdown stop
        return time

    @property
    def rf_ramp_voltages(self):
        """List of voltages to define RF ramp."""
        vals = (self.rf_ramp_bottom_voltage,
                self.rf_ramp_bottom_voltage,
                self.rf_ramp_top_voltage,
                self.rf_ramp_top_voltage,
                self.rf_ramp_bottom_voltage)
        return vals

    @property
    def rf_ramp_phases(self):
        """List of phases to define RF ramp."""
        p = (self.rf_ramp_bottom_phase,
             self.rf_ramp_bottom_phase,
             self.rf_ramp_top_phase,
             self.rf_ramp_top_phase,
             self.rf_ramp_bottom_phase)
        return p

    def rf_ramp_interp_voltages(self, time):
        """Return voltages related to times."""
        vals = _np.interp(time, self.rf_ramp_times, self.rf_ramp_voltages)
        return vals

    @property
    def rf_ramp_rampup_start_time(self):
        """RF ramp rampup start time."""
        time = self.ti_params_rf_ramp_delay + \
            float(self._value['rf_ramp']['bottom_duration'])
        return time

    @rf_ramp_rampup_start_time.setter
    def rf_ramp_rampup_start_time(self, value):
        value = float(value)
        if value == self.rf_ramp_rampup_start_time:
            return
        if not self.ti_params_rf_ramp_delay <= value < \
                self.rf_ramp_rampup_stop_time:
            raise _RampInvalidRFParms('Invalid rampup start time.')

        delay = self.ti_params_rf_ramp_delay
        rampup_stop_time = self.rf_ramp_rampup_stop_time
        self._value['rf_ramp']['bottom_duration'] = \
            value - delay
        self._value['rf_ramp']['rampup_duration'] = \
            rampup_stop_time - value
        self.verify_synchronized()

    @property
    def rf_ramp_rampup_stop_time(self):
        """RF ramp rampup stop time."""
        time = self.ti_params_rf_ramp_delay + \
            float(self._value['rf_ramp']['bottom_duration']) + \
            float(self._value['rf_ramp']['rampup_duration'])
        return time

    @rf_ramp_rampup_stop_time.setter
    def rf_ramp_rampup_stop_time(self, value):
        value = float(value)
        if value == self.rf_ramp_rampup_stop_time:
            return
        if not self.rf_ramp_rampup_start_time < value <= \
                self.rf_ramp_rampdown_start_time:
            raise _RampInvalidRFParms('Invalid rampup stop time.')

        rampup_start_time = self.rf_ramp_rampup_start_time
        rampdown_start_time = self.rf_ramp_rampdown_start_time
        self._value['rf_ramp']['rampup_duration'] = \
            value - rampup_start_time
        self._value['rf_ramp']['top_duration'] = \
            rampdown_start_time - value
        self.verify_synchronized()

    @property
    def rf_ramp_rampdown_start_time(self):
        """RF ramp rampdown start time."""
        time = self.ti_params_rf_ramp_delay + \
            float(self._value['rf_ramp']['bottom_duration']) + \
            float(self._value['rf_ramp']['rampup_duration']) + \
            float(self._value['rf_ramp']['top_duration'])
        return time

    @rf_ramp_rampdown_start_time.setter
    def rf_ramp_rampdown_start_time(self, value):
        value = float(value)
        if value == self.rf_ramp_rampdown_start_time:
            return
        if not self.rf_ramp_rampup_stop_time <= value < \
                self.rf_ramp_rampdown_stop_time:
            raise _RampInvalidRFParms('Invalid rampdown start time.')

        rampup_stop_time = self.rf_ramp_rampup_stop_time
        rampdown_stop_time = self.rf_ramp_rampdown_stop_time
        self._value['rf_ramp']['top_duration'] = \
            value - rampup_stop_time
        self._value['rf_ramp']['rampdown_duration'] = \
            rampdown_stop_time - value
        self.verify_synchronized()

    @property
    def rf_ramp_rampdown_stop_time(self):
        """RF ramp rampdown start time."""
        time = self.ti_params_rf_ramp_delay + \
            float(self._value['rf_ramp']['bottom_duration']) + \
            float(self._value['rf_ramp']['rampup_duration']) + \
            float(self._value['rf_ramp']['top_duration']) + \
            float(self._value['rf_ramp']['rampdown_duration'])
        return time

    @rf_ramp_rampdown_stop_time.setter
    def rf_ramp_rampdown_stop_time(self, value):
        value = float(value)
        if value == self.rf_ramp_rampdown_stop_time:
            return
        if not self.rf_ramp_rampdown_start_time < value <= \
                self.ti_params_rf_ramp_delay + _MAX_RF_RAMP_DURATION:
            raise _RampInvalidRFParms('Invalid rampdown stop time.')

        rampdown_start_time = self.rf_ramp_rampdown_start_time
        self._value['rf_ramp']['rampdown_duration'] = \
            value - rampdown_start_time
        self.verify_synchronized()

    @property
    def rf_ramp_bottom_duration(self):
        """Bottom duration, in ms."""
        return self._value['rf_ramp']['bottom_duration']

    @property
    def rf_ramp_rampup_duration(self):
        """Rampup duration, in ms."""
        return self._value['rf_ramp']['rampup_duration']

    @property
    def rf_ramp_top_duration(self):
        """Top duration, in ms."""
        return self._value['rf_ramp']['top_duration']

    @property
    def rf_ramp_rampdown_duration(self):
        """Rampdown duration, in ms."""
        return self._value['rf_ramp']['rampdown_duration']

    @property
    def rf_ramp_bottom_voltage(self):
        """RF ramp bottom voltage, in kV."""
        return self._value['rf_ramp']['bottom_voltage']

    @rf_ramp_bottom_voltage.setter
    def rf_ramp_bottom_voltage(self, value):
        value = float(value)
        if value == self._value['rf_ramp']['bottom_voltage']:
            return
        if not 0 <= value <= self.rf_ramp_top_voltage:
            raise _RampInvalidRFParms('Invalid value to bottom voltage.')

        self._value['rf_ramp']['bottom_voltage'] = value
        self.verify_synchronized()

    @property
    def rf_ramp_top_voltage(self):
        """RF ramp top voltage, in kV."""
        return self._value['rf_ramp']['top_voltage']

    @rf_ramp_top_voltage.setter
    def rf_ramp_top_voltage(self, value):
        value = float(value)
        if value == self._value['rf_ramp']['top_voltage']:
            return
        if not value >= self._value['rf_ramp']['bottom_voltage']:
            raise _RampInvalidRFParms('Invalid value to top voltage.')

        self._value['rf_ramp']['top_voltage'] = value
        self.verify_synchronized()

    @property
    def rf_ramp_bottom_phase(self):
        """RF ramp bottom phase, in degrees."""
        return self._value['rf_ramp']['bottom_phase']

    @rf_ramp_bottom_phase.setter
    def rf_ramp_bottom_phase(self, value):
        value = float(value)
        if value == self._value['rf_ramp']['bottom_phase']:
            return
        if not -180 < value < 360:
            raise _RampInvalidRFParms('Invalid value to bottom phase.')

        self._value['rf_ramp']['bottom_phase'] = value
        self.verify_synchronized()

    @property
    def rf_ramp_top_phase(self):
        """RF ramp top phase, in degrees."""
        return self._value['rf_ramp']['top_phase']

    @rf_ramp_top_phase.setter
    def rf_ramp_top_phase(self, value):
        value = float(value)
        if value == self._value['rf_ramp']['top_phase']:
            return
        if not -180 < value < 360:
            raise _RampInvalidRFParms('Invalid value to top phase.')

        self._value['rf_ramp']['top_phase'] = value
        self.verify_synchronized()

    # --- timing parameters ---

    @property
    def ti_params_injection_time(self):
        """Injection time instant."""
        return float(self._value['ti_params']['injection_time'])

    @ti_params_injection_time.setter
    def ti_params_injection_time(self, value):
        """Set injection time instant."""
        # TODO: verify value
        if value == self._value['ti_params']['injection_time']:
            return
        self._value['ti_params']['injection_time'] = value
        self.verify_synchronized()

    @property
    def ti_params_ejection_time(self):
        """Ejection time instant."""
        return float(self._value['ti_params']['ejection_time'])

    @ti_params_ejection_time.setter
    def ti_params_ejection_time(self, value):
        """Set ejection time instant."""
        # TODO: verify value
        if value == self._value['ti_params']['ejection_time']:
            return
        self._value['ti_params']['ejection_time'] = value
        self.verify_synchronized()

    @property
    def ti_params_ps_ramp_delay(self):
        """PS ramp delay."""
        return float(self._value['ti_params']['ps_ramp_delay'])

    @ti_params_ps_ramp_delay.setter
    def ti_params_ps_ramp_delay(self, value):
        """Set ps ramp delay [us]."""
        value = float(value)
        if value == self._value['ti_params']['ps_ramp_delay']:
            return
        self._value['ti_params']['ps_ramp_delay'] = value
        self.verify_synchronized()

    @property
    def ti_params_rf_ramp_delay(self):
        """RF delay."""
        return float(self._value['ti_params']['rf_ramp_delay'])

    @ti_params_rf_ramp_delay.setter
    def ti_params_rf_ramp_delay(self, value):
        """Set RF ramp delay [us]."""
        # TODO: verify value
        if value == self._value['ti_params']['rf_ramp_delay']:
            return
        self._value['ti_params']['rf_ramp_delay'] = value
        self.verify_synchronized()

    # --- API for waveforms ---

    @property
    def ps_waveform_psnames_exclimits(self):
        """Return a list of psnames whose waveform exceeds current limits."""
        if not self._value['ps_normalized_configs*']:
            psnames = list(self.PSNAME_DIPOLES)
        else:
            psnames = self.PSNAMES
        psnames_exclimits = list()
        for psname in psnames:
            self._update_ps_waveform(psname)
            w_currents = self._ps_waveforms[psname].currents
            isnan = _np.any(_np.isnan(w_currents))
            isinf = _np.any(_np.isinf(w_currents))
            if isnan or isinf:
                continue
            limits = _PSSearch.conv_psname_2_splims(psname)
            highlim = limits['HOPR'] if psname not in self.PSNAME_DIPOLES \
                else 1072
            lowlim = limits['LOPR']
            if _np.any(w_currents > highlim) or _np.any(w_currents < lowlim):
                psnames_exclimits.append(psname)
        return psnames_exclimits

    @property
    def ps_waveform_psnames_init_end_diff(self):
        if not self._value['ps_normalized_configs*']:
            return list()
        psnames = _dcopy(self.PSNAMES)
        for psn in self.PSNAMES:
            if psn in self.PSNAME_DIPOLES:
                psnames.remove(psn)

        psnames_initenddiff = list()
        for psname in psnames:
            self._update_ps_waveform(psname)
            w_currents = self._ps_waveforms[psname].currents
            c_init = w_currents[0]
            c_end = w_currents[-1]
            if not _np.isclose(c_init, c_end):
                psnames_initenddiff.append(psname)
        return psnames_initenddiff

    def ps_waveform_get(self, psname):
        """Return ps waveform for a given power supply."""
        if psname not in self.PSNAME_DIPOLES and \
                not self._value['ps_normalized_configs*']:
            raise _RampError('There is no normalized cofiguration defined!')
        self._update_ps_waveform(psname)
        waveform = self._ps_waveforms[psname]
        return waveform

    def ps_waveform_set(self, psname, waveform):
        """Set ps waveform for a given power supply."""
        self._ps_waveforms[psname] = _dcopy(waveform)

    def ps_waveform_get_times(self, psname):
        """Return ramp energy at a given time."""
        self._update_ps_waveform(psname)
        times = self._ps_waveforms[psname].times
        return times

    def ps_waveform_get_currents(self, psname):
        """Return ps waveform current for a given power supply."""
        if not self._value['ps_normalized_configs*'] and \
                psname not in self.PSNAME_DIPOLES:
            raise _RampError('There is no normalized cofiguration defined!')
        self._update_ps_waveform(psname)
        waveform = self._ps_waveforms[psname]
        return waveform.currents.copy()

    def ps_waveform_get_strengths(self, psname):
        """Return ps waveform strength for a given power supply."""
        if not self._value['ps_normalized_configs*'] and \
                psname not in self.PSNAME_DIPOLES:
            raise _RampError('There is no normalized cofiguration defined!')
        self._update_ps_waveform(psname)
        waveform = self._ps_waveforms[psname]
        return waveform.strengths.copy()

    def ps_waveform_interp_time(self, energy):
        """Return ps ramp time at a given energy.
        Use only energies until rampdown-start time.
        """
        rampdown_start_time = self.ps_ramp_rampdown_start_time
        times = [time for time in
                 self.ps_waveform_get_times(self.PSNAME_DIPOLE_REF)
                 if time < rampdown_start_time]
        energies = self._ps_waveforms[
            self.PSNAME_DIPOLE_REF].strengths[0:len(times)]
        time = _np.interp(energy, energies, times)
        return time

    def ps_waveform_interp_strengths(self, psname, time):
        """Return ps ramp strength at a given time."""
        if not self._value['ps_normalized_configs*'] and \
                psname not in self.PSNAME_DIPOLES:
            raise _RampError('There is no normalized cofiguration defined!')
        self._update_ps_waveform(psname)
        times = self.ps_waveform_get_times(psname)
        strengths = self._ps_waveforms[psname].strengths
        strength = _np.interp(time, times, strengths)
        return strength

    def ps_waveform_interp_currents(self, psname, time):
        """Return ps ramp current at a given time."""
        if not self._value['ps_normalized_configs*'] and \
                psname not in self.PSNAME_DIPOLES:
            raise _RampError('There is no normalized cofiguration defined!')
        self._update_ps_waveform(psname)
        times = self.ps_waveform_get_times(psname)
        currents = self._ps_waveforms[psname].currents
        current = _np.interp(time, times, currents)
        return current

    def ps_waveform_interp_energy(self, time):
        """Return ps ramp energy at a given time."""
        return self.ps_waveform_interp_strengths(self.PSNAME_DIPOLE_REF, time)

    def verify_synchronized(self):
        """."""
        if not self._orig_value:
            # print('no original value to compare')
            return
        modified = False
        for group, params in self._orig_value.items():
            if not params:
                if self._value[group]:
                    modified = True
                    # print('now, it has configs', self._value[group].keys())
                    break
                else:
                    # print('do not have configs')
                    continue
            if self._orig_value[group].keys() != self._value[group].keys():
                modified = True
                # print('has different keys',
                #       self._orig_value[group].keys(),
                #       self._value[group].keys())
                break
            for param, value in params.items():
                if self._value[group][param] != value:
                    modified = True
                    # print('has params with different values',
                    #       self._value[group][param], value)
                    break
        self._synchronized = not modified

    def verify_ps_normalized_synchronized(self, time, value=None, prec=1e-6):
        """."""
        str_time = '{:.3f}'.format(time)
        if not self._orig_value:
            return False
        if str_time not in self._orig_value['ps_normalized_configs*'].keys():
            return False

        o = self._orig_value['ps_normalized_configs*'][str_time]
        c = value if value is not None else \
            self._value['ps_normalized_configs*'][str_time]

        diff = list()
        for val in c:
            if val == 'label':
                if c[val] != o[val]:
                    diff.append((val, c[val], o[val]))
            else:
                if not _np.isclose(c[val], o[val], atol=prec):
                    diff.append((val, c[val], o[val]))
        # print(diff)
        return len(diff) == 0

    # --- private methods ---

    def __len__(self):
        """Return number of ps normalized configurations."""
        return len(self._value['ps_normalized_configs*'])

    def __str__(self):
        """Return string representation of configuration."""
        if not self._value:
            rst = 'name: {}'.format(self.name)
            return rst
        labels = (
            'ti_params_rf_ramp_delay [us]',
            'ti_params_ps_ramp_delay [us]',
            'ti_params_injection_time [ms]',
            'ti_params_ejection_time [ms]',
            'ps_ramp_duration [ms]',
            'ps_ramp_time_energy [ms] [GeV]',
            'ps_ramp_slopes [GeV/s]',
            'ps_normalized_configs [ms] [name]',
        )
        rst = ''
        maxlen = max(tuple(len(l) for l in labels) + (len('name'),))
        strfmt1 = '{:<' + str(maxlen) + 's}: {}\n'
        strfmt2 = strfmt1.replace('{}', '{:07.3f} {:+08.3f} {:<s}')
        strfmt3 = strfmt1.replace('{}', '{:07.3f} {}')
        strfmt4 = strfmt1.replace('{}', '{:07.3f}')
        rst += strfmt1.format('name', self.name)
        rst += strfmt1.format(labels[0], self.ti_params_rf_ramp_delay)
        rst += strfmt1.format(labels[1], self.ti_params_ps_ramp_delay)
        rst += strfmt4.format(labels[2], self.ti_params_injection_time)
        rst += strfmt4.format(labels[3], self.ti_params_ejection_time)
        rst += strfmt1.format(labels[4], self.ps_ramp_duration)
        rst += strfmt1.format(labels[5], '')
        rst += strfmt2.format(
            '', 0.0, self.ps_ramp_start_energy, '(start)')
        rst += strfmt2.format(
            '', self.ps_ramp_rampup1_start_time,
            self.ps_ramp_rampup1_start_energy, '(rampup1_start)')
        rst += strfmt2.format(
            '', self.ps_ramp_rampup2_start_time,
            self.ps_ramp_rampup2_start_energy, '(rampup2_start)')
        rst += strfmt2.format(
            '', self.ps_ramp_rampdown_start_time,
            self.ps_ramp_rampdown_start_energy, '(rampdown_start)')
        rst += strfmt2.format(
            '', self.ps_ramp_rampdown_stop_time,
            self.ps_ramp_rampdown_stop_energy, '(rampdown_stop)')
        rst += strfmt2.format(
            '', self.ps_ramp_duration,
            self.ps_ramp_start_energy, '(stop)')
        rst += strfmt1.format(labels[6], '')
        rst += strfmt3.format(
            '', self.ps_ramp_rampup1_slope, '(rampup1)')
        rst += strfmt3.format('', self.ps_ramp_rampup2_slope, '(rampup2)')
        rst += strfmt3.format(
            '', self.ps_ramp_rampdown_slope, '(rampdown)')
        rst += strfmt1.format(labels[7], '')
        time = self.ps_normalized_configs_times
        name = self.ps_normalized_configs_labels
        for i, _time in enumerate(time):
            rst += strfmt3.format('', _time, name[i])
        return rst

    def _get_item(self, time):
        str_time = '{:.3f}'.format(time)
        return _dcopy(self._value['ps_normalized_configs*'][str_time])

    def _set_item(self, time, value):
        """Set configuration item."""
        str_time = '{:.3f}'.format(time)
        if value == self._value['ps_normalized_configs*'][str_time]:
            return
        self._value['ps_normalized_configs*'][str_time] = _dcopy(value)
        self.verify_synchronized()
        self._invalidate_ps_waveforms()

    def _set_value(self, value):
        super()._set_value(value)
        self.verify_synchronized()
        self._invalidate_ps_waveforms()

    def _update_ps_normalized_config_energy(self, nconfig_obj, time):
        indices = self._conv_times_2_indices(self.PSNAME_DIPOLE_REF, [time])
        strengths = self.ps_waveform_get_strengths(self.PSNAME_DIPOLE_REF)
        strength = _np.interp(
            indices[0],
            list(range(self._get_appropriate_wfmnrpoints(
                self.PSNAME_DIPOLE_REF))),
            strengths)
        for psname in self.PSNAME_DIPOLES:
            nconfig_obj[psname] = strength
        return nconfig_obj

    def _update_ps_waveform(self, psname):

        # update dipole if necessary
        if self.PSNAME_DIPOLE_REF not in self._ps_waveforms:
            self._update_ps_waveform_dipole()

        # update family if necessary
        family = _get_magnet_family_name(psname)
        if family is not None and family not in self._ps_waveforms:
            self._update_ps_waveform(family)

        # update magnet waveform if it is not a dipole
        if psname not in self._ps_waveforms:
            self._update_ps_waveform_not_dipole(
                psname, self.PSNAME_DIPOLE_REF, family)

    def _update_ps_waveform_not_dipole(self, psname, dipole, family=None):
        nc_times = sorted(self.ps_normalized_configs_times)
        nc_values = list()
        for time in nc_times:
            nconfig = self._value['ps_normalized_configs*'][
                '{:.3f}'.format(time)]
            nc_values.append(nconfig[psname])

        # interpolate strengths
        wfm_nrpoints = self._get_appropriate_wfmnrpoints(psname)
        nc_indices = self._conv_times_2_indices(psname, nc_times)
        wfm_indices = [i for i in range(wfm_nrpoints)]
        wfm_strengths = _np.interp(wfm_indices, nc_indices, nc_values)

        # create waveform object with given strengths
        dipole = self._ps_waveforms[dipole]
        if family is not None:
            family = self._ps_waveforms[family]
        self._ps_waveforms[psname] = _Waveform(
            psname=psname, dipole=dipole, family=family,
            strengths=wfm_strengths, wfm_nrpoints=wfm_nrpoints)

    def _update_ps_waveform_dipole(self):
        for psname in self.PSNAME_DIPOLES:
            wfm = self._create_new_ps_waveform_dipole(psname)
            self._ps_waveforms[psname] = wfm

    def _create_new_ps_waveform_dipole(self, dipole=''):
        if not dipole:
            dipole = self.PSNAME_DIPOLE_REF
        rdip = self._value['ps_ramp']
        dipole = _WaveformDipole(
            psname=dipole,
            wfm_nrpoints=self._get_appropriate_wfmnrpoints(dipole),
            duration=rdip['duration'],
            start_energy=rdip['start_energy'],
            rampup1_start_time=rdip['rampup1_start_time'],
            rampup1_start_energy=rdip['rampup1_start_energy'],
            rampup2_start_time=rdip['rampup2_start_time'],
            rampup2_start_energy=rdip['rampup2_start_energy'],
            rampup_smooth_intvl=rdip['rampup_smooth_intvl'],
            rampup_smooth_energy=rdip['rampup_smooth_energy'],
            rampdown_start_time=rdip['rampdown_start_time'],
            rampdown_start_energy=rdip['rampdown_start_energy'],
            rampdown_stop_time=rdip['rampdown_stop_time'],
            rampdown_stop_energy=rdip['rampdown_stop_energy'],
            rampdown_smooth_intvl=rdip['rampdown_smooth_intvl'],
            rampdown_smooth_energy=rdip['rampdown_smooth_energy'])
        return dipole

    def _conv_times_2_indices(self, psname, times):
        duration = self.ps_ramp_duration
        wfm_nrpoints = self._get_appropriate_wfmnrpoints(psname)
        interval = duration / (wfm_nrpoints - 1.0)
        indices = [time/interval for time in times]
        return indices

    def _invalidate_ps_waveforms(self, include_dipole=False):
        psnames = tuple(self._ps_waveforms.keys())
        for psname in psnames:
            if psname not in self.PSNAME_DIPOLES or include_dipole:
                del self._ps_waveforms[psname]

    def _check_valid_time(self, time):
        """Check if time is in valid interval."""
        duration = self.ps_ramp_duration
        if 0 < time < duration:
            return True
        else:
            raise _RampError(
                'Time value must be between 0 and {}!'.format(duration))

    def _get_appropriate_wfmnrpoints(self, psname):
        """Return appropriate number of points for psname."""
        if _PSSearch.conv_psname_2_psmodel(psname) == 'FBP':
            return self.ps_ramp_wfm_nrpoints_corrs
        else:
            return self.ps_ramp_wfm_nrpoints_fams


class SiriusMig(BoosterRamp):
    """Sirius migration class."""

    PSNAME_DIPOLES = ('SI-Fam:PS-B1B2-1', 'SI-Fam:PS-B1B2-2')



"""Devices subpackage."""

from .afc_acq_core import AFCPhysicalTrigger, AFCACQLogicalTrigger
from .bbb import BunchbyBunch
from .bpm import BPM, FamBPMs
from .currinfo import CurrInfoTranspEff, CurrInfoLinear, \
    CurrInfoBO, CurrInfoSI, CurrInfoAS
from .dcct import DCCT
from .device import Device, DeviceApp, Devices, DeviceNC
from .egun import EGBias, EGFilament, EGHVPS, EGTriggerPS, EGPulsePS, EGun
from .energy import Energy
from .fofb import FOFBCtrlDCC, BPMDCC, FOFBCtrlRef, FamFOFBControllers, \
    FamFastCorrs, HLFOFB
from .fofb_acq import FOFBCtrlSysId, FOFBPSSysId, FamFOFBSysId, \
    FOFBCtrlLamp, FOFBPSLamp, FamFOFBLamp
from .ict import ICT, TranspEff
from .ids import APU, WIG, PAPU, EPU
from .idff import IDFF, WIGIDFF, PAPUIDFF, EPUIDFF, APUIDFF
from .injctrl import InjCtrl
from .injsys import PUMagsStandbyHandler, BOPSRampStandbyHandler, \
    BORFRampStandbyHandler, InjSysStandbyHandler, LinacStandbyHandler, \
    InjSysPUModeHandler
from .lillrf import LILLRF, DevLILLRF
from .machshift import MachShift
from .modltr import LIModltr
from .orbit_interlock import BPMOrbitIntlk, BaseOrbitIntlk, OrbitInterlock
from .posang import PosAng
from .psconv import PSProperty, StrengthConv
from .pssofb import PSCorrSOFB, PSApplySOFB
from .pwrsupply import PowerSupply, PowerSupplyPU, PowerSupplyFC, \
    PowerSupplyFBP
from .rf import RFGen, ASLLRF, BORFCavMonitor, SIRFCavMonitor, RFCav, \
    RFKillBeam
from .screen import Screen
from .sofb import SOFB
from .syncd import DevicesSync
from .timing import EVG, Event, Trigger, HLTiming
from .tune import TuneFrac, TuneProc, Tune, TuneCorr
from .dvf import DVF, DVFImgProc
from .lienergy import LIEnergy
from .blctrl import BLPPSCtrl
from .scraper import ScraperH, ScraperV


del device, bpm, dcct, egun, ict, lillrf, modltr
del pwrsupply, posang, psconv, pssofb, rf, injsys, injctrl
del screen, tune, sofb, timing, syncd, energy
del ids, currinfo, bbb, machshift, dvf, lienergy, blctrl, scraper



#!/usr/bin/env python-sirius

"""Siriuspy setup script."""

import pkg_resources
from setuptools import find_packages, setup


def get_abs_path(relative):
    return pkg_resources.resource_filename(__name__, relative)


with open(get_abs_path("README.md"), "r") as _f:
    _long_description = _f.read().strip()

with open(get_abs_path("VERSION"), "r") as _f:
    __version__ = _f.read().strip()

with open(get_abs_path("requirements.txt"), "r") as _f:
    _requirements = _f.read().strip().split("\n")

setup(
    name="siriuspy",
    version=__version__,
    author="lnls-sirius",
    description="Development packages for Sirius",
    long_description=_long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/lnls-sirius/dev-packages",
    download_url="https://github.com/lnls-sirius/dev-packages",
    license="GNU GPLv3",
    classifiers=[
        "Intended Audience :: Science/Research",
        "Programming Language :: Python",
        "Topic :: Scientific/Engineering",
    ],
    packages=find_packages(),
    install_requires=_requirements,
    package_data={"siriuspy": ["VERSION"]},
    include_package_data=True,
    test_suite="tests",
    python_requires=">=3.6",
    zip_safe=False,
)



"""SI chromaticity correction configuration.

Values in _template_dict are arbitrary. They are used just to compare with
corresponding values when a new configuration is tried to be inserted in the
servconf database.
"""
from copy import deepcopy as _dcopy


def get_dict():
    """Return configuration type dictionary."""
    module_name = __name__.split('.')[-1]
    _dict = {
        'config_type_name': module_name,
        'value': _dcopy(_template_dict)
    }
    return _dict


# Chromaticity Correction Parameters for Storage Ring
#
# | ChromX |   | NominalChromX |   | m00  m01...m014 |   | SL SI-Fam:PS-SFA1 |
# |        | = |               | + |                 | * |         .         |
# | ChromY |   | NominalChromY |   | m10  m11...m114 |   |         .         |
#                                                        |         .         |
#                                                        | SL SI-Fam:PS-SDP3 |
# Data ordering:
#
# Nominal Chromaticity
#    NominalChromX    NominalChromY
#
# Correction Matrix of Additional Method
# (obtained from matlab lnls_calc_chromcorr_params routine)
#   m(0,0)   m(0,1) ... m(0,14)
#   m(1,0)   m(1,1) ... m(1,14)
#
# Nominals SLs
# [sextupole_order  SFA1  SFA2  SDA1  SDA2  SDA3  SFB1  SFB2  SDB1  SDB2  SDB3  SFP1  SFP2  SDP1  SDP2  SDP3]


_template_dict = {
    'nominal chrom': [0.0, 0.0],
    'matrix': [[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
               [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]],
    'nominal SLs': [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                    0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
}



"""Define PVs, contants and properties of High Level FOFB."""

import os as _os
import numpy as _np

from .. import csdev as _csdev
from ..search import PSSearch as _PSSearch, MASearch as _MASearch, \
    BPMSearch as _BPMSearch


NR_BPM = 160


# --- Enumeration Types ---

class ETypes(_csdev.ETypes):
    """Local enumerate types."""

    OPEN_CLOSED = ('Open', 'Closed')

    GLOB_INDIV = ('Global', 'Individual')

    MEAS_RMAT_CMD = ('Start', 'Stop', 'Reset')
    MEAS_RMAT_MON = ('Idle', 'Measuring', 'Completed', 'Aborted')

    STS_LBLS_CORR = (
        'Connected', 'PwrStateOn', 'OpModeConfigured', 'AccFreezeConfigured',
        'InvRespMatRowSynced', 'AccGainSynced', 'AccSatLimsSynced',
        'AccDecimationSynced')
    STS_LBLS_FOFBCTRL = (
        'Connected', 'BPMIdsConfigured', 'NetSynced', 'LinkPartnerConnected',
        'RefOrbSynced', 'TimeFrameLenSynced', 'BPMLogTrigsConfigured',
        'OrbDistortionDetectionSynced', 'PacketLossDetectionSynced',
        'LoopInterlockOk', 'SYSIDExcitationDisabled')

    DEC_OPT = ('FOFB', 'Monit', 'Custom')


_et = ETypes  # syntactic sugar


# --- Const class ---

class HLFOFBConst(_csdev.Const):
    """Const class defining High Level FOFB constants."""

    MIN_SING_VAL = 0.1
    TIKHONOV_REG_CONST = 0
    SINGVALHW_THRS = 1e-14
    DEF_MAX_ORB_DISTORTION = 200  # [um]

    CONV_UM_2_NM = 1e3
    ACCGAIN_RESO = 2**-12

    DEF_TIMEOUT = 10  # [s]
    DEF_TIMESLEEP = 0.1  # [s]
    DEF_TIMEWAIT = 3  # [s]
    DEF_TIMEMINWAIT = 1  # [s]
    LOOPGAIN_RMP_TIME = 5  # [s]
    LOOPGAIN_RMP_FREQ = 2  # [steps/s]
    LOOPGAIN_RMP_NPTS = LOOPGAIN_RMP_TIME * LOOPGAIN_RMP_FREQ
    CURRZERO_RMP_FREQ = 2  # [steps/s]

    LoopState = _csdev.Const.register('LoopState', _et.OPEN_CLOSED)
    GlobIndiv = _csdev.Const.register('GlobIndiv', _et.GLOB_INDIV)
    UseRF = _csdev.Const.register('UseRF', _et.DSBL_ENBL)
    MeasRespMatCmd = _csdev.Const.register('MeasRespMatCmd', _et.MEAS_RMAT_CMD)
    MeasRespMatMon = _csdev.Const.register('MeasRespMatMon', _et.MEAS_RMAT_MON)
    DecOpt = _csdev.Const.register('DecOpt', _et.DEC_OPT)

    def __init__(self):
        """Class constructor."""

        # device names and nicknames
        self.bpm_names = _BPMSearch.get_names({'sec': 'SI', 'dev': 'BPM'})
        if NR_BPM != len(self.bpm_names):
            raise ValueError('Inconsistent NR_BPM parameter!')
        self.ch_names = _PSSearch.get_psnames({'sec': 'SI', 'dev': 'FCH'})
        self.cv_names = _PSSearch.get_psnames({'sec': 'SI', 'dev': 'FCV'})

        self.bpm_nicknames = _BPMSearch.get_nicknames(self.bpm_names)
        self.ch_nicknames = _PSSearch.get_psnicknames(self.ch_names)
        self.cv_nicknames = _PSSearch.get_psnicknames(self.cv_names)

        # device position along the ring
        self.bpm_pos = _BPMSearch.get_positions(self.bpm_names)
        self.ch_pos = _MASearch.get_mapositions(map(
            lambda x: x.substitute(dis='MA'), self.ch_names))
        self.cv_pos = _MASearch.get_mapositions(map(
            lambda x: x.substitute(dis='MA'), self.cv_names))

        # total number of devices
        self.nr_bpms = NR_BPM
        self.nr_ch = len(self.ch_names)
        self.nr_cv = len(self.cv_names)
        self.nr_chcv = self.nr_ch + self.nr_cv
        self.nr_corrs = self.nr_chcv + 1  # include RF

        # data folder
        ioc_fol = _os.path.join(
            '/home', 'sirius', 'iocs-log', 'si-ap-fofb', 'data')
        self.reforb_fname = _os.path.join(ioc_fol, 'reforbit.orb')
        self.respmat_fname = _os.path.join(ioc_fol, 'respmat.respmat')
        self.bpmxenbl_fname = _os.path.join(ioc_fol, 'bpmxenbllist.bpmenbl')
        self.bpmyenbl_fname = _os.path.join(ioc_fol, 'bpmyenbllist.bpmenbl')
        self.chenbl_fname = _os.path.join(ioc_fol, 'chenbllist.correnbl')
        self.cvenbl_fname = _os.path.join(ioc_fol, 'cvenbllist.correnbl')
        self.rfenbl_fname = _os.path.join(ioc_fol, 'rfenbllist.rfenbl')

        # reforb and matrix parameters
        self.reforb_size = self.nr_bpms
        self.matrix_size = self.nr_corrs * (2 * self.nr_bpms)
        self.nr_svals = min(self.nr_corrs, 2 * self.nr_bpms)
        self.corrcoeffs_size = self.nr_chcv * (2 * self.nr_bpms)
        self.corrgains_size = self.nr_chcv

        # dcc minimum enable configuration
        self.dccenbl_min = _np.array([
            bpm.sub[2:] in ['M1', 'M2'] for bpm in self.bpm_names])

    def get_hlfofb_database(self):
        """Return Soft IOC database."""
        pvs_database = {
            # Global
            'Version-Cte': {'type': 'string', 'value': 'UNDEF'},
            'Log-Mon': {'type': 'string', 'value': 'Starting...'},

            'LoopState-Sel': {
                'type': 'enum', 'enums': _et.OPEN_CLOSED,
                'value': self.LoopState.Open},
            'LoopState-Sts': {
                'type': 'enum', 'enums': _et.OPEN_CLOSED,
                'value': self.LoopState.Open},
            'LoopGainH-SP': {
                'type': 'float', 'value': 0.0520, 'prec': 4,
                'lolim': -2**3, 'hilim': 2**3-1,
                'unit': 'FOFB pre-accumulator gain.'},
            'LoopGainH-RB': {
                'type': 'float', 'value': 0.0520, 'prec': 4,
                'lolim': -2**3, 'hilim': 2**3-1,
                'unit': 'FOFB pre-accumulator gain.'},
            'LoopGainH-Mon': {
                'type': 'float', 'value': 0, 'prec': 4,
                'unit': 'FOFB pre-accumulator gain.'},
            'LoopGainV-SP': {
                'type': 'float', 'value': 0.0520, 'prec': 4,
                'lolim': -2**3, 'hilim': 2**3-1,
                'unit': 'FOFB pre-accumulator gain.'},
            'LoopGainV-RB': {
                'type': 'float', 'value': 0.0520, 'prec': 4,
                'lolim': -2**3, 'hilim': 2**3-1,
                'unit': 'FOFB pre-accumulator gain.'},
            'LoopGainV-Mon': {
                'type': 'float', 'value': 0, 'prec': 4,
                'unit': 'FOFB pre-accumulator gain.'},
            'LoopMaxOrbDistortion-SP': {
                'type': 'float', 'value': self.DEF_MAX_ORB_DISTORTION,
                'prec': 3, 'unit': 'um', 'lolim': 0, 'hilim': 10000},
            'LoopMaxOrbDistortion-RB': {
                'type': 'float', 'value': self.DEF_MAX_ORB_DISTORTION,
                'prec': 3, 'unit': 'um', 'lolim': 0, 'hilim': 10000},
            'LoopMaxOrbDistortionEnbl-Sel': {
                'type': 'enum', 'enums': _et.DSBL_ENBL,
                'value': self.DsblEnbl.Dsbl},
            'LoopMaxOrbDistortionEnbl-Sts': {
                'type': 'enum', 'enums': _et.DSBL_ENBL,
                'value': self.DsblEnbl.Dsbl},
            'LoopPacketLossDetecEnbl-Sel': {
                'type': 'enum', 'enums': _et.DSBL_ENBL,
                'value': self.DsblEnbl.Dsbl},
            'LoopPacketLossDetecEnbl-Sts': {
                'type': 'enum', 'enums': _et.DSBL_ENBL,
                'value': self.DsblEnbl.Dsbl},

            # Correctors
            'CHPosS-Cte': {
                'type': 'float', 'unit': 'm', 'count': self.nr_ch,
                'value': self.ch_pos},
            'CVPosS-Cte': {
                'type': 'float', 'unit': 'm', 'count': self.nr_cv,
                'value': self.cv_pos},
            'CHNickName-Cte': {
                'type': 'string', 'unit': 'shortname for the chs.',
                'count': self.nr_ch, 'value': self.ch_nicknames},
            'CVNickName-Cte': {
                'type': 'string', 'unit': 'shortname for the cvs.',
                'count': self.nr_cv, 'value': self.cv_nicknames},
            'CorrStatus-Mon': {'type': 'int', 'value': 0b11111111},
            'CorrStatusLabels-Cte': {
                'type': 'string', 'count': len(_et.STS_LBLS_CORR),
                'value': _et.STS_LBLS_CORR},
            'CorrConfig-Cmd': {'type': 'int', 'value': 0},
            'CorrSetPwrStateOn-Cmd': {'type': 'int', 'value': 0},
            'CorrSetPwrStateOff-Cmd': {'type': 'int', 'value': 0},
            'CorrSetOpModeManual-Cmd': {'type': 'int', 'value': 0},
            'CorrSetAccFreezeDsbl-Cmd': {'type': 'int', 'value': 0},
            'CorrSetAccFreezeEnbl-Cmd': {'type': 'int', 'value': 0},
            'CorrSetAccClear-Cmd': {'type': 'int', 'value': 0},
            'CorrSetCurrZero-Cmd': {'type': 'int', 'value': 0},
            'CorrSetCurrZeroDuration-SP': {
                'type': 'float', 'prec': 0, 'value': 5, 'unit': 's',
                'lolim': 0.0, 'hilim': 1000.0},
            'CorrSetCurrZeroDuration-RB': {
                'type': 'float', 'prec': 0, 'value': 5, 'unit': 's',
                'lolim': 0.0, 'hilim': 1000.0},
            'CHAccSatMax-SP': {
                'type': 'float', 'prec': 6, 'value': 0.3, 'unit': 'A',
                'lolim': 0, 'hilim': 0.95},
            'CHAccSatMax-RB': {
                'type': 'float', 'prec': 6, 'value': 0.3, 'unit': 'A',
                'lolim': 0, 'hilim': 0.95},
            'CVAccSatMax-SP': {
                'type': 'float', 'prec': 6, 'value': 0.3, 'unit': 'A',
                'lolim': 0, 'hilim': 0.95},
            'CVAccSatMax-RB': {
                'type': 'float', 'prec': 6, 'value': 0.3, 'unit': 'A',
                'lolim': 0, 'hilim': 0.95},

            # FOFB Controllers
            'TimeFrameLen-SP': {
                'type': 'int', 'value': 2100, 'lolim': 500, 'hilim': 10000},
            'TimeFrameLen-RB': {
                'type': 'int', 'value': 2100, 'lolim': 500, 'hilim': 10000},
            'CtrlrStatus-Mon': {'type': 'int', 'value': 0b11111111111},
            'CtrlrStatusLabels-Cte': {
                'type': 'string', 'count': len(_et.STS_LBLS_FOFBCTRL),
                'value': _et.STS_LBLS_FOFBCTRL},
            'CtrlrConfBPMId-Cmd': {'type': 'int', 'value': 0},
            'CtrlrSyncNet-Cmd': {'type': 'int', 'value': 0},
            'CtrlrSyncUseEnblList-Sel': {
                'type': 'enum', 'enums': _et.DSBL_ENBL,
                'value': self.DsblEnbl.Enbl},
            'CtrlrSyncUseEnblList-Sts': {
                'type': 'enum', 'enums': _et.DSBL_ENBL,
                'value': self.DsblEnbl.Enbl},
            'CtrlrSyncEnblList-Mon': {
                'type': 'int', 'count': self.nr_bpms,
                'value': self.nr_bpms*[1], 'unit': 'DCCs used in loop'},
            'CtrlrSyncRefOrb-Cmd': {'type': 'int', 'value': 0},
            'CtrlrSyncTFrameLen-Cmd': {'type': 'int', 'value': 0},
            'CtrlrConfBPMLogTrg-Cmd': {'type': 'int', 'value': 0},
            'CtrlrSyncMaxOrbDist-Cmd': {'type': 'int', 'value': 0},
            'CtrlrSyncPacketLossDetec-Cmd': {'type': 'int', 'value': 0},
            'CtrlrReset-Cmd': {'type': 'int', 'value': 0},
            'CtrlrDsblSYSIDExc-Cmd': {'type': 'int', 'value': 0},

            # decimation configuration
            'FOFBAccDecimation-Sel': {
                'type': 'enum', 'enums': _et.DEC_OPT,
                'value': self.DecOpt.FOFB, 'unit': 'FOFB_Monit_Custom'},
            'FOFBAccDecimation-Sts': {
                'type': 'enum', 'enums': _et.DEC_OPT,
                'value': self.DecOpt.FOFB, 'unit': 'FOFB_Monit_Custom'},
            'FOFBAccDecimation-SP': {
                'type': 'float', 'value': 1, 'prec': 0, 'lolim': 1,
                'hilim': 8600, 'unit': 'count'},
            'FOFBAccDecimation-RB': {
                'type': 'float', 'value': 1, 'prec': 0, 'lolim': 1,
                'hilim': 8600, 'unit': 'count'},

            # Reference Orbit (same order of SOFB)
            'RefOrbX-SP': {
                'type': 'float', 'unit': 'um', 'count': self.nr_bpms,
                'value': self.nr_bpms*[0]},
            'RefOrbX-RB': {
                'type': 'float', 'unit': 'um', 'count': self.nr_bpms,
                'value': self.nr_bpms*[0]},
            'RefOrbY-SP': {
                'type': 'float', 'unit': 'um', 'count': self.nr_bpms,
                'value': self.nr_bpms*[0]},
            'RefOrbY-RB': {
                'type': 'float', 'unit': 'um', 'count': self.nr_bpms,
                'value': self.nr_bpms*[0]},
            'RefOrbHwX-Mon': {
                'type': 'int', 'unit': 'nm', 'count': self.nr_bpms,
                'value': self.nr_bpms*[0]},
            'RefOrbHwY-Mon': {
                'type': 'int', 'unit': 'nm', 'count': self.nr_bpms,
                'value': self.nr_bpms*[0]},

            # Response Matrix
            'BPMXEnblList-SP': {
                'type': 'int', 'count': self.nr_bpms,
                'value': self.nr_bpms*[1], 'unit': 'BPMX used in correction'},
            'BPMXEnblList-RB': {
                'type': 'int', 'count': self.nr_bpms,
                'value': self.nr_bpms*[1], 'unit': 'BPMX used in correction'},
            'BPMYEnblList-SP': {
                'type': 'int', 'count': self.nr_bpms,
                'value': self.nr_bpms*[1], 'unit': 'BPMY used in correction'},
            'BPMYEnblList-RB': {
                'type': 'int', 'count': self.nr_bpms,
                'value': self.nr_bpms*[1], 'unit': 'BPMY used in correction'},
            'CHEnblList-SP': {
                'type': 'int', 'count': self.nr_ch, 'value': self.nr_ch*[1],
                'unit': 'CHs used in correction'},
            'CHEnblList-RB': {
                'type': 'int', 'count': self.nr_ch, 'value': self.nr_ch*[1],
                'unit': 'CHs used in correction'},
            'CVEnblList-SP': {
                'type': 'int', 'count': self.nr_cv, 'value': self.nr_cv*[1],
                'unit': 'CVs used in correction'},
            'CVEnblList-RB': {
                'type': 'int', 'count': self.nr_cv, 'value': self.nr_cv*[1],
                'unit': 'CVs used in correction'},
            'UseRF-Sel': {
                'type': 'enum', 'enums': _et.DSBL_ENBL,
                'value': self.UseRF.Enbl,
                'unit': 'If RF is used in inverse matrix calculation'},
            'UseRF-Sts': {
                'type': 'enum', 'enums': _et.DSBL_ENBL,
                'value': self.UseRF.Enbl,
                'unit': 'If RF is used in inverse matrix calculation'},
            'MinSingValue-SP': {
                'type': 'float', 'value': self.MIN_SING_VAL, 'prec': 5,
                'lolim': 0, 'hilim': 1e9,
                'unit': 'Maximum value of SV to use'},
            'MinSingValue-RB': {
                'type': 'float', 'value': self.MIN_SING_VAL, 'prec': 5,
                'lolim': 0, 'hilim': 1e9,
                'unit': 'Maximum value of SV to use'},
            'TikhonovRegConst-SP': {
                'type': 'float', 'value': self.TIKHONOV_REG_CONST, 'prec': 5,
                'lolim': 0, 'hilim': 1e9,
                'unit': 'Tikhonov regularization constant'},
            'TikhonovRegConst-RB': {
                'type': 'float', 'value': self.TIKHONOV_REG_CONST, 'prec': 5,
                'lolim': 0, 'hilim': 1e9,
                'unit': 'Tikhonov regularization constant'},
            'SingValuesRaw-Mon': {
                'type': 'float', 'count': self.nr_svals,
                'value': self.nr_svals*[0],
                'unit': 'Singular values of the matrix'},
            'SingValues-Mon': {
                'type': 'float', 'count': self.nr_svals,
                'value': self.nr_svals*[0],
                'unit': 'Singular values of the matrix in use'},
            'NrSingValues-Mon': {
                'type': 'int', 'value': self.nr_svals,
                'lolim': 1, 'hilim': self.nr_svals,
                'unit': 'Number of used SVs'},
            'RespMat-SP': {
                'type': 'float', 'count': self.matrix_size,
                'value': self.matrix_size*[0],
                'unit': '(BH, BV)(nm) x (CH, CV, RF)(A, Hz)'},
            'RespMat-RB': {
                'type': 'float', 'count': self.matrix_size,
                'value': self.matrix_size*[0],
                'unit': '(BH, BV)(nm) x (CH, CV, RF)(A, Hz)'},
            'RespMat-Mon': {
                'type': 'float', 'count': self.matrix_size,
                'value': self.matrix_size*[0],
                'unit': '(BH, BV)(um) x (CH, CV, RF)(urad, Hz)'},
            'InvRespMat-Mon': {
                'type': 'float', 'count': self.matrix_size,
                'value': self.matrix_size*[0],
                'unit': '(CH, CV, RF)(urad, Hz) x (BH, BV)(um)'},
            'SingValuesHw-Mon': {
                'type': 'float', 'count': self.nr_svals,
                'value': self.nr_svals*[0],
                'unit': 'Singular values of the matrix in hardware units'},
            'RespMatHw-Mon': {
                'type': 'float', 'count': self.matrix_size,
                'value': self.matrix_size*[0],
                'unit': '(BH, BV)(nm) x (CH, CV, RF)(counts, Hz)'},
            'InvRespMatHw-Mon': {
                'type': 'float', 'count': self.matrix_size,
                'value': self.matrix_size*[0],
                'unit': '(CH, CV, RF)(counts, Hz) x (BH, BV)(nm)'},
            'InvRespMatNormMode-Sel': {
                'type': 'enum', 'enums': _et.GLOB_INDIV,
                'value': self.GlobIndiv.Global},
            'InvRespMatNormMode-Sts': {
                'type': 'enum', 'enums': _et.GLOB_INDIV,
                'value': self.GlobIndiv.Global},
            'CorrCoeffs-Mon': {
                'type': 'float', 'count': self.corrcoeffs_size,
                'value': self.corrcoeffs_size*[0],
                'unit': '(CH, CV) x (BH, BV)'},
            'CorrGains-Mon': {
                'type': 'float', 'count': self.corrgains_size,
                'value': self.corrgains_size*[0],
                'unit': '(CH, CV)'},

            # Response Matrix Measurement
            'MeasRespMat-Cmd': {
                'type': 'enum', 'enums': _et.MEAS_RMAT_CMD,
                'value': self.MeasRespMatCmd.Start},
            'MeasRespMat-Mon': {
                'type': 'enum', 'enums': _et.MEAS_RMAT_MON,
                'value': self.MeasRespMatMon.Idle},
            'MeasRespMatKickCH-SP': {
                'type': 'float', 'value': 15, 'unit': 'urad', 'prec': 3,
                'lolim': 0.002, 'hilim': 30},
            'MeasRespMatKickCH-RB': {
                'type': 'float', 'value': 15, 'unit': 'urad', 'prec': 3,
                'lolim': 0.002, 'hilim': 30},
            'MeasRespMatKickCV-SP': {
                'type': 'float', 'value': 22.5, 'unit': 'urad', 'prec': 3,
                'lolim': 0.002, 'hilim': 30},
            'MeasRespMatKickCV-RB': {
                'type': 'float', 'value': 22.5, 'unit': 'urad', 'prec': 3,
                'lolim': 0.002, 'hilim': 30},
            'MeasRespMatKickRF-SP': {
                'type': 'float', 'value': 75, 'unit': 'Hz', 'prec': 2,
                'lolim': 1, 'hilim': 1000},
            'MeasRespMatKickRF-RB': {
                'type': 'float', 'value': 75, 'unit': 'Hz', 'prec': 2,
                'lolim': 1, 'hilim': 1000},
            'MeasRespMatWait-SP': {
                'type': 'float', 'value': 1, 'unit': 's', 'prec': 3,
                'lolim': 0.005, 'hilim': 100},
            'MeasRespMatWait-RB': {
                'type': 'float', 'value': 1, 'unit': 's', 'prec': 3,
                'lolim': 0.005, 'hilim': 100},
        }
        pvs_database = _csdev.add_pvslist_cte(pvs_database)
        return pvs_database



"""Machine shift utils."""

import os as _os

import numpy as _np


# This solution is a simplified version of scipy.interpolate.interp1d for
# interpolation of kind 'previous' with fill_value='extrapolate' option
def interp1d_previous(x_org, y_org, x_new):
    """interp1d to previous."""
    x_new = _np.asarray(x_new)
    x_org = _np.asarray(x_org).ravel()
    y_org = _np.asarray(y_org)

    # Get index of left value
    x_new_indices = _np.searchsorted(
        _np.nextafter(x_org, -_np.inf), x_new, side='left')

    # Clip x_new_indices so that they are within the range of x_org indices.
    x_new_indices = x_new_indices.clip(1, len(x_org)).astype(_np.intp)

    # Calculate the actual value for each entry in x_new.
    y_new = y_org[x_new_indices-1]

    return y_new


# Version using scipy.interpolate.interp1d
# from scipy.interpolate import interp1d as _interp1d
# def _interp1d_previous(x_org, y_org, x_new):
#     """interp1d to previous."""
#     fun = _interp1d(x_org, y_org, 'previous', fill_value='extrapolate')
#     y_new = fun(x_new)
#     return y_new


class HandleMachShiftFile:
    """Class to handle machine shift file."""

    def __init__(self):
        """Init."""
        self.fpath = '/home/sirius/iocs-log/as-ap-machshift/'
        self.fname = self.fpath + 'last-mach-shift.txt'

    def get_machshift(self):
        """Get last shift from file."""
        if _os.path.isfile(self.fname):
            with open(self.fname, 'r') as fil:
                machshift = fil.read().strip('\n')
        else:
            _os.makedirs(self.fpath, exist_ok=True)
            machshift = 1  # Commissioning
            with open(self.fname, 'w+') as fil:
                fil.write(machshift)
        return int(machshift)

    def set_machshift(self, machshift):
        """Set last shift in file."""
        if not isinstance(machshift, int):
            raise TypeError("'machshift' must be an int.")
        with open(self.fname, 'w+') as fil:
            fil.write(str(machshift))
